
__author__    = "Vivek Balasubramanian <vivek.balasubramanian@rutgers.edu>"
__copyright__ = "Copyright 2017, http://radical.rutgers.edu"
__license__   = "MIT"


# ------------------------------------------------------------------------------
#
class EnTKError(Exception):
    '''
    EnTKError is the generic exception type used by EnTK -- exception arg
    messages are usually
    '''
    pass


# ------------------------------------------------------------------------------
#
class EnTKTypeError(TypeError):
    '''
    TypeError is raised if value of a wrong type is passed to a function or
    assigned as an attribute of an object
    '''

    def __init__(self, expected_type, actual_type, entity=None):

        if entity:
            msg = "Entity: %s, Expected (base) type(s) %s, but got %s." \
                  % (str(entity), str(expected_type), str(actual_type))
        else:
            msg = "Expected (base) type(s) %s, but got %s." \
                  % (str(expected_type), str(actual_type))

        super(EnTKTypeError, self).__init__(msg)


# ------------------------------------------------------------------------------
#
class EnTKValueError(ValueError):
    '''
    ValueError is raised if a value that is unacceptable is passed to a
    function or assigned as an attribute of an object
    '''

    def __init__(self, obj, attribute, expected_value, actual_value):

        if type(expected_value) != list:
            msg = 'Value for attribute %s of object %s incorrect. ' \
                  'Expected value %s, but got %s.' % (str(attribute),
                               str(obj), str(expected_value), str(actual_value))
        else:
            text = ','.join([str(s) for s in expected_value])
            msg  = 'Value for attribute %s of object %s incorrect. ' \
                   'Expected values %s, but got %s.' % (str(attribute),
                                        str(obj), text, str(actual_value))

        super().__init__(msg)


# ------------------------------------------------------------------------------
#
class EnTKMissingError(AttributeError):
    '''
    MissingError is raised when an attribute that is mandatory is left
    unassigned by the user
    '''

    def __init__(self, obj, missing_attribute):

        super().__init__('Attribute %s in %s undefined'
                         % (str(missing_attribute), str(obj)))

# ------------------------------------------------------------------------------





import os
import pickle

import radical.pilot as rp
import radical.utils as ru

from ...task       import Task
from ...exceptions import EnTKTypeError, EnTKValueError


# ------------------------------------------------------------------------------
#
def resolve_placeholders(path, placeholders, logger):
    """
    **Purpose**: Substitute placeholders in staging attributes of a Task with
                 actual paths to the corresponding tasks.

    :arguments:
        :path:             string describing the staging paths, possibly
                           containing a placeholder
        :placeholders: dictionary holding the values for placeholders
    """

    try:

        if isinstance(path, str):
            path = str(path)

        if not isinstance(path, str):
            raise EnTKTypeError(expected_type=str, actual_type=type(path))

        if '$' not in path:
            return path

        # Extract placeholder from path
        path_placeholders = list()
        if len(path.split('>')) == 1:
            path_placeholders.append(path.split('/')[0])

        else:
            if path.split('>')[0].strip().startswith('$'):
                path_placeholders.append(
                        path.split('>')[0].strip().split('/')[0])

            if path.split('>')[1].strip().startswith('$'):
                path_placeholders.append(
                        path.split('>')[1].strip().split('/')[0])

        resolved = path
        for placeholder in path_placeholders:

            # SHARED
            if placeholder == "$SHARED":
                resolved = resolved.replace(placeholder, 'pilot://')
                continue

            # Expected placeholder format:
            # $Pipeline_{pipeline.uid}_Stage_{stage.uid}_Task_{task.uid}
            elems = placeholder.split('/')[0].split('_')
            if not len(elems) == 6:

                expected = '$Pipeline_(pipeline_name)_' \
                           'Stage_(stage_name)_' \
                           'Task_(task_name) or $SHARED',
                raise EnTKValueError(obj='placeholder', attribute='task',
                                    expected_value=expected, actual_value=elems)

            pname    = elems[1]
            sname    = elems[3]
            tname    = elems[5]
            is_resolved_by_uid = False

            if pname in placeholders:
                if sname in placeholders[pname]:
                    if tname in placeholders[pname][sname]:
                        resolved = resolved.replace(placeholder,
                                   placeholders[pname][sname][tname]['path'])
                        is_resolved_by_uid = True
                    else:
                        logger.warning('%s not assigned to any task in '
                                'Stage %s Pipeline %s' % (tname, sname, pname))
                else:
                    logger.warning('%s not assigned to any Stage in Pipeline %s'
                                  % (sname, pname))
            else:
                logger.warning('%s not assigned to any Pipeline' % (pname))

            if is_resolved_by_uid is False:
                placeholders_by_name = placeholders["__by_name__"]
                if pname in placeholders_by_name:
                    if sname in placeholders_by_name[pname]:
                        if tname in placeholders_by_name[pname][sname]:
                            resolved = resolved.replace(placeholder,
                              placeholders_by_name[pname][sname][tname]['path'])
                        else:
                            logger.warning('%s not assigned to any task in '
                                 'Stage %s Pipeline %s' % (tname, sname, pname))
                    else:
                        logger.warning('%s not assigned to any Stage in '
                                       'Pipeline %s' % (sname, pname))
                else:
                    logger.warning('%s not assigned to any Pipeline' % (pname))


            if not resolved:
                logger.warning('No placeholder could be found for task name %s \
                            stage name %s and pipeline name %s. Please be sure \
                            to use object names and not uids in your \
                            references,i.e, \
                 $Pipeline_(pipeline_name)_Stage_(stage_name)_Task_(task_name)')
                expected = '$Pipeline_(pipeline_name)_' \
                           'Stage_(stage_name)_' \
                           'Task_(task_name) or $SHARED'
                raise EnTKValueError(obj='placeholder', attribute='task',
                                    expected_value=expected, actual_value=elems)

        return resolved

    except Exception as ex:

        logger.exception('Failed to resolve placeholder %s, error: %s' %
                         (path, ex))
        raise


# ------------------------------------------------------------------------------
#
def resolve_arguments(args, placeholders, logger):

    resolved_args = list()

    for entry in args:

        # If entry starts with $, it has a placeholder
        # and needs to be resolved based after a lookup in
        # the placeholders
        if not isinstance(entry, str) or \
           not entry.startswith('$'):

            resolved_args.append(entry)
            continue


        placeholder = entry.split('/')[0]

        if placeholder == "$SHARED":
            entry = entry.replace(placeholder, '$RP_PILOT_SANDBOX')

        elif placeholder.startswith('$Pipeline'):
            elems = placeholder.split('_')

            if len(elems) != 6:

                expected = '$Pipeline_{pipeline.uid}_' \
                           'Stage_{stage.uid}_' \
                           'Task_{task.uid} or $SHARED'
                raise EnTKValueError(obj='placeholder', attribute='length',
                                    expected_value=expected, actual_value=elems)

            pname = elems[1]
            sname = elems[3]
            tname = elems[5]

            try:
                entry = entry.replace(placeholder,
                                      placeholders[pname][sname][tname]['path'])

            except Exception:
                logger.warning('Argument parsing failed. Task %s of Stage %s '
                               'in Pipeline %s does not exist',
                               tname, sname, pname)

        resolved_args.append(entry)

    return resolved_args


# ------------------------------------------------------------------------------
#
def resolve_tags(task, parent_pipeline_name, placeholders):

    tags = dict(task.tags or {'colocate': task.uid})

    pnames = list(placeholders)
    pnames.remove(parent_pipeline_name)

    colo_tag = tags['colocate']
    for pname in [parent_pipeline_name] + pnames:
        for sname in placeholders[pname]:
            if colo_tag in placeholders[pname][sname]:
                tags['colocate'] = placeholders[pname][sname][colo_tag]['uid']
                return tags

    return tags


# ------------------------------------------------------------------------------
#
def get_input_list_from_task(task, placeholders, logger):
    """
    Purpose: Parse Task object to extract the files to be staged as input.

    Details: The extracted data is then converted into the appropriate RP
             directive depending on whether the data is to be copied/downloaded.

    :arguments:
        :task:         EnTK Task object
        :placeholders: dictionary holding the values for placeholders

    :return: list of RP directives for the files that need to be staged in
    """

    try:

        if not isinstance(task, Task):
            raise EnTKTypeError(expected_type=Task, actual_type=type(task))

        input_data = list()

        if task.link_input_data:

            for path in task.link_input_data:

                path = resolve_placeholders(path, placeholders, logger)

                if len(path.split('>')) > 1:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.LINK
                    }

                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.LINK
                    }
                input_data.append(temp)


        if task.upload_input_data:

            for path in task.upload_input_data:

                path = resolve_placeholders(path, placeholders, logger)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip()
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip())
                    }
                input_data.append(temp)


        if task.copy_input_data:

            for path in task.copy_input_data:

                path = resolve_placeholders(path, placeholders, logger)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.COPY
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.COPY
                    }
                input_data.append(temp)


        if task.move_input_data:

            for path in task.move_input_data:

                path = resolve_placeholders(path, placeholders, logger)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.MOVE
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.MOVE
                    }

                input_data.append(temp)

        return input_data


    except Exception:

        logger.exception('Failed to get input list of files from task')
        raise


# ------------------------------------------------------------------------------
#
def get_output_list_from_task(task, placeholders, logger):
    """
    Purpose: Parse Task object to extract the files to be staged as the output.

    Details: The extracted data is then converted into the appropriate RP
             directive depending on whether the data is to be copied/downloaded.

    :arguments:
        :task:         EnTK Task object
        :placeholders: dictionary holding the values for placeholders

    :return: list of RP directives for the files that need to be staged out

    """

    try:

        if not isinstance(task, Task):
            raise EnTKTypeError(expected_type=Task, actual_type=type(task))


        output_data = list()

        if task.link_output_data:

            for path in task.link_output_data:

                path = resolve_placeholders(path, placeholders, logger)

                if len(path.split('>')) > 1:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.LINK
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.LINK
                    }
                output_data.append(temp)

        if task.download_output_data:

            for path in task.download_output_data:

                path = resolve_placeholders(path, placeholders, logger)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip()
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip())
                    }
                output_data.append(temp)

        if task.copy_output_data:

            for path in task.copy_output_data:

                path = resolve_placeholders(path, placeholders, logger)

                if len(path.split('>')) > 1:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.COPY
                    }

                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.COPY
                    }
                output_data.append(temp)

        if task.move_output_data:

            for path in task.move_output_data:

                path = resolve_placeholders(path, placeholders, logger)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.MOVE
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.MOVE
                    }

                output_data.append(temp)

        return output_data


    except Exception:
        logger.exception('Failed to get output list of files from task')
        raise


# ------------------------------------------------------------------------------
#
def create_td_from_task(task, placeholders, task_hash_table, pkl_path, sid,
                        logger, prof=None):
    """
    Purpose: Create an RP Task description based on the defined Task.

    :arguments:
        :task:         EnTK Task object
        :placeholders: dictionary holding the values for placeholders

    :return: rp.TaskDescription
    """

    try:

        logger.debug('Creating Task from Task %s: %s' % (task.uid, task.sandbox))
        logger.debug('Hash table state: %s' % task_hash_table)

        if prof:
            prof.prof('td_create', uid=task.uid)

        td = rp.TaskDescription()

        task_pre_uid = task_hash_table.get(task.uid, None)
        if task_pre_uid is None:
            td.uid = task.uid
        else:
            tmp_uid = ru.generate_id(prefix=task.uid, ns=sid)
            td.uid = tmp_uid
        task_hash_table[task.uid] = td.uid
        with open(pkl_path, 'wb') as pickle_file:
            pickle.dump(task_hash_table, pickle_file, pickle.HIGHEST_PROTOCOL)

        logger.debug('Hash table state: %s' % task_hash_table)

        td.name = '%s,%s,%s,%s,%s,%s' % (task.uid, task.name,
                                         task.parent_stage['uid'],
                                         task.parent_stage['name'],
                                         task.parent_pipeline['uid'],
                                         task.parent_pipeline['name'])

        td.pre_launch     = task.pre_launch
        td.pre_exec       = task.pre_exec
        td.executable     = task.executable
        td.arguments      = resolve_arguments(task.arguments, placeholders, logger)
        td.sandbox        = task.sandbox
        td.post_exec      = task.post_exec
        td.environment    = task.environment
        td.post_launch    = task.post_launch
        td.stage_on_error = task.stage_on_error

        if task.annotations:
            td.metadata   = {'data': task.annotations.as_dict()}

        if task.parent_pipeline['uid']:
            td.tags = resolve_tags(
                task=task,
                parent_pipeline_name=task.parent_pipeline['uid'],
                placeholders=placeholders)

        td.ranks          = task.cpu_reqs['cpu_processes']
        td.cores_per_rank = task.cpu_reqs['cpu_threads']
        td.gpus_per_rank  = task.gpu_reqs['gpu_processes']

        if task.cpu_reqs['cpu_thread_type']:
            td.threading_type = task.cpu_reqs['cpu_thread_type']
        else:
            td.threading_type = rp.POSIX

        if task.gpu_reqs['gpu_process_type']:
            td.gpu_type = task.gpu_reqs['gpu_process_type']
        else:
            td.gpu_type = ''

        if task.lfs_per_process:
            td.lfs_per_process = task.lfs_per_process

        if task.mem_per_process:
            td.mem_per_process = task.mem_per_process

        if task.stdout: td.stdout = task.stdout
        if task.stderr: td.stderr = task.stderr

        td.input_staging  = get_input_list_from_task(task, placeholders, logger)
        td.output_staging = get_output_list_from_task(task, placeholders, logger)

        if prof:
            prof.prof('td from task - done', uid=task.uid)

        logger.debug('Task %s created from Task %s' % (td.name, task.uid))

        return td

    except Exception:
        logger.exception('Task creation failed')
        raise


# ------------------------------------------------------------------------------
#
def create_task_from_rp(rp_task, logger, prof=None):
    """
    Purpose: Create a Task based on the RP Task.

    Details: Currently, only the uid, parent_stage and parent_pipeline are
             retrieved. The exact initial Task (that was converted to a TD)
             cannot be recovered as the RP API does not provide the same
             attributes for a Task as for a TD.  Also, this is not required for
             the most part.

    :arguments:
        :task: RP Task

    :return: Task
    """

    try:
        logger.debug('Create Task from RP Task %s' % rp_task.name)

        if prof:
            prof.prof('task_create', uid=rp_task.name.split(',')[0].strip())

        task_info_tuple = rp_task.name.split(',')
        task = Task({
            'uid'                     : task_info_tuple[0].strip(),
            'name'                    : task_info_tuple[1].strip(),
            'parent_stage'   : {'uid' : task_info_tuple[2].strip(),
                                'name': task_info_tuple[3].strip()},
            'parent_pipeline': {'uid' : task_info_tuple[4].strip(),
                                'name': task_info_tuple[5].strip()},
            'rts_uid'                 : rp_task.uid
        })

        if   rp_task.state == rp.DONE                  : task.exit_code = 0
        elif rp_task.state in [rp.FAILED, rp.CANCELED] : task.exit_code = 1

        if rp_task.state == rp.FAILED:
            task.exception        = rp_task.exception
            task.exception_detail = rp_task.exception_detail

        task.path = ru.Url(rp_task.sandbox).path

        if prof:
            prof.prof('task_created', uid=task.uid)

        logger.debug('Task %s created from RP Task %s' % (task.uid, rp_task.name))

        return task


    except Exception:
        logger.exception('Task creation from RP Task failed, error')
        raise


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2014-2020, http://radical.rutgers.edu'
__license__   = 'MIT'

import threading

import radical.utils as ru

from string      import punctuation

from .constants  import NAME_MESSAGE
from .exceptions import EnTKError, EnTKMissingError
from .exceptions import EnTKTypeError, EnTKValueError
from .stage      import Stage
from .           import states


class Pipeline(object):

    """
    A pipeline represents a collection of objects that have a linear temporal
    execution order.  In this case, a pipeline consists of multiple 'Stage'
    objects. Each ```Stage_i``` can execute only after all stages up to
    ```Stage_(i-1)``` have completed execution.

    """

    def __init__(self):

        self._uid  = ru.generate_id('pipeline.%(counter)04d', ru.ID_CUSTOM)
        self._name = None

        self._stages = list()
        self._state  = states.INITIAL

        # Keep track of states attained
        self._state_history = [states.INITIAL]

        # To keep track of current state
        self._stage_count = len(self._stages)
        self._cur_stage   = 0

        # Lock around current stage
        self._lock = threading.Lock()

        # To keep track of termination of pipeline
        self._completed_flag = threading.Event()


    # --------------------------------------------------------------------------
    # Getter functions
    # --------------------------------------------------------------------------

    @property
    def name(self):
        """
        Name of the pipeline useful for bookkeeping and to refer to this
        pipeline while data staging. Do not use a ',' or '_' in an object's name.

        :getter: Returns the name of the pipeline
        :setter: Assigns the name of the pipeline
        :type: String
        """
        return self._name

    @property
    def stages(self):
        """
        Stages of the list

        :getter: Returns the stages in the current Pipeline
        :setter: Assigns the stages to the current Pipeline
        :type: List
        """

        return self._stages

    @property
    def state(self):
        """
        Current state of the pipeline

        :getter: Returns the state of the current pipeline
        :type: String
        """

        return self._state

    @property
    def uid(self):
        """
        Unique ID of the current pipeline

        :getter: Returns the unique id of the current pipeline
        :type: String
        """
        return self._uid


    @property
    def luid(self):
        """
        Unique ID of the current pipeline (fully qualified).
        For the pipeline class, his is an alias to `uid`.

        example:
            >>> pipeline.luid
            pipe.0001

        :getter: Returns the fully qualified uid of the current pipeline
        :type: String
        """
        if self.name: return self.name
        else        : return self.uid


    @property
    def lock(self):
        """
        Returns the lock over the current Pipeline

        :return: Lock object
        """

        return self._lock

    @property
    def completed(self):
        """
        Returns whether the Pipeline has completed

        :return: Boolean
        """
        return self._completed_flag.is_set()

    @property
    def current_stage(self):
        """
        Returns the current stage being executed

        :return: Integer
        """

        return self._cur_stage

    @property
    def state_history(self):
        """
        Returns a list of the states obtained in temporal order

        :return: list
        """

        return self._state_history

    # --------------------------------------------------------------------------
    # Setter functions
    # --------------------------------------------------------------------------

    @name.setter
    def name(self, value):

        invalid_symbols = punctuation.replace('.','')
        if not isinstance(value, str):
            raise EnTKTypeError(expected_type=str,
                                actual_type=type(value))

        if any(symbol in value for symbol in invalid_symbols):
            raise EnTKValueError(obj=self._uid,
                                 attribute='name',
                                 actual_value=value,
                                 expected_value=NAME_MESSAGE)

        self._name = value

    @stages.setter
    def stages(self, value):

        self._stages = self._validate_entities(value)

        self._stage_count = len(self._stages)
        if self._cur_stage == 0:
            self._cur_stage = 1

    @state.setter
    def state(self, value):
        if isinstance(value, str):
            if value in list(states._pipeline_state_values.keys()):  # pylint: disable=W0212
                self._state = value

                # We add SUSPENDED to state history in suspend()
                if self._state != states.SUSPENDED:
                    self._state_history.append(value)
            else:
                raise EnTKValueError(obj=self._uid, attribute='state',
                      expected_value=list(states._pipeline_state_values.keys()),  # pylint: disable=W0212
                                 actual_value=value)
        else:
            raise EnTKTypeError(expected_type=str, actual_type=type(value))


    # --------------------------------------------------------------------------
    #
    def add_stages(self, value):
        """
        Appends stages to the current Pipeline

        :argument: List of Stage objects
        """
        stages = self._validate_entities(value)

        self._stages.extend(stages)
        self._stage_count = len(self._stages)
        if self._cur_stage == 0:
            self._cur_stage = 1

        for stage in stages:
            stage.parent_pipeline['uid']  = self._uid
            stage.parent_pipeline['name'] = self._name

            for task in stage.tasks:
                task.parent_pipeline['uid']  = self._uid
                task.parent_pipeline['name'] = self._name


    def as_dict(self):
        """
        Convert current Pipeline (i.e. its attributes) into a dictionary

        :return: python dictionary
        """

        pipeline_desc_as_dict = {

            'uid': self._uid,
            'name': self._name,
            'state': self._state,
            'state_history': self._state_history,
            'completed': self._completed_flag.is_set()
        }

        return pipeline_desc_as_dict

    def from_dict(self, d):
        """
        Create a Pipeline from a dictionary. The change is in inplace.

        :argument: python dictionary
        :return: None
        """

        if 'uid' in d:
            if d['uid']:
                self._uid = d['uid']

        if 'name' in d:
            if d['name']:
                invalid_symbols = punctuation.replace('.','')
                if not isinstance(d['name'], str):
                    raise EnTKTypeError(expected_type=str,
                                        actual_type=type(d['name']))

                if any(symbol in d['name'] for symbol in invalid_symbols):
                    raise EnTKValueError(obj=self._uid,
                                        attribute='name',
                                        actual_value=d['name'],
                                        expected_value=NAME_MESSAGE)

                self._name = d['name']

        if 'state' in d:
            if isinstance(d['state'], str) or isinstance(d['state'], str):
                if d['state'] in list(states._pipeline_state_values.keys()):  # pylint: disable=W0212
                    self._state = d['state']
                else:
                    raise EnTKValueError(obj=self._uid, attribute='state',
                      expected_value=list(states._pipeline_state_values.keys()),  # pylint: disable=W0212
                                     actual_value=d['state'])
            else:
                raise EnTKTypeError(entity='state', expected_type=str,
                                actual_type=type(d['state']))

        else:
            self._state = states.INITIAL

        if 'state_history' in d:
            if isinstance(d['state_history'], list):
                self._state_history = d['state_history']
            else:
                raise EnTKTypeError(entity='state_history', expected_type=list,
                        actual_type=type(d['state_history']))

        if 'completed' in d:
            if isinstance(d['completed'], bool):
                if d['completed']:
                    self._completed_flag.set()
            else:
                raise EnTKTypeError(entity='completed', expected_type=bool,
                                actual_type=type(d['completed']))

    # --------------------------------------------------------------------------
    #
    def suspend(self):
        '''
        Pause execution of the pipeline: stages and tasks that are executing
        will continue to execute, but no new stages and tasks will be eligible
        for execution until `resume()` is called.

         - The `suspend()` method can not be called on a suspended or completed
           pipeline, doing so will result in an exeption.
         - The state of the pipeline will be set to `SUSPENDED`.
        '''
        if self._state == states.SUSPENDED:
            raise EnTKError('suspend() called on suspended Pipeline %s'
                            % self._uid)

        self._state = states.SUSPENDED
        self._state_history.append(self._state)


    # --------------------------------------------------------------------------
    #
    def resume(self):
        '''
        Continue execution of paused stages and tasks.

         - The `resume()` method can only be called on a suspended pipeline, an
           exception will be raised if that condition is not met.
         - The state of a resumed pipeline will be set to the state the pipeline
           had before suspension.
        '''
        if self._state != states.SUSPENDED:
            raise EnTKError('Cannot resume Pipeline %s: not suspended [%s] [%s]'
                    % (self._uid, self._state, self._state_history))

        self._state = self._state_history[-2]
        self._state_history.append(self._state)


    # --------------------------------------------------------------------------
    # Private methods
    # --------------------------------------------------------------------------

    def _increment_stage(self):
        """
        Purpose: Increment stage pointer. Also check if Pipeline has completed.
        """

        try:

            if self._cur_stage < self._stage_count:
                self._cur_stage += 1
            else:
                self._completed_flag.set()

        except Exception as ex:
            raise EnTKError(ex) from ex

    def _decrement_stage(self):
        """
        Purpose: Decrement stage pointer. Reset completed flag.
        """

        try:

            if self._cur_stage > 0:
                self._cur_stage -= 1
                self._completed_flag = threading.Event()  # reset

        except Exception as ex:
            raise EnTKError(ex) from ex

    @classmethod
    def _validate_entities(self, stages):
        """
        Purpose: Validate whether the argument 'stages' is of list of
        Stage objects

        :argument: list of Stage objects
        """
        if not stages:
            raise EnTKTypeError(expected_type=Stage, actual_type=type(stages))

        if not isinstance(stages, list):
            stages = [stages]

        for value in stages:
            if not isinstance(value, Stage):
                raise EnTKTypeError(expected_type=Stage, actual_type=type(value))

        return stages

    def _validate(self):
        """
        Purpose: Validate that the state of the current Pipeline is 'DESCRIBED'
        (user has not meddled with it). Also validate that the current Pipeline
        contains Stages.
        """

        if self._state is not states.INITIAL:

            raise EnTKValueError(obj=self._uid, attribute='state',
                        expected_value=states.INITIAL, actual_value=self._state)

        if not self._stages:

            raise EnTKMissingError(obj=self._uid, missing_attribute='stages')

        annotated_outputs = {}
        for stage in self._stages:
            stage._validate()  # pylint: disable=W0212

            # validate provided annotations
            for task in stage.tasks:

                if not task.annotations:
                    continue

                annotated_outputs[task.uid] = task.annotations.outputs
                for t_input in task.annotations.inputs:
                    delimiter_count = t_input.count(':')
                    if not delimiter_count:
                        continue
                    elif delimiter_count == 1 and '://' in t_input:
                        # TODO: extend file representation format
                        continue
                    # ensure that task inputs are annotated as outputs
                    # for corresponding tasks it depends on
                    t_dep_uid, t_dep_output = t_input.split(':', maxsplit=1)
                    if t_dep_output not in annotated_outputs.get(t_dep_uid, []):
                        raise EnTKError(
                            'Annotation error for %s: '      % task.uid +
                            'provided input "%s" '           % t_dep_output +
                            'is not produced by assigned %s' % t_dep_uid)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__    = "Vivek Balasubramanian <vivek.balasubramanian@rutgers.edu>"
__license__   = "MIT"


import os

import radical.utils as ru

from ...exceptions import EnTKError, EnTKTypeError

from .resource_manager import Base_ResourceManager


# ------------------------------------------------------------------------------
#
class Base_TaskManager(object):
    """
    A Task Manager takes the responsibility of dispatching tasks it receives
    from a 'pending' queue for execution on to the available resources using a
    runtime system. Once the tasks have completed execution, they are pushed
    on to the completed queue for other components of EnTK to thread.

    :arguments:
        :rmgr:              (ResourceManager) Object to be used to access the
                            Pilot where the tasks can be submitted

    Currently, EnTK is configured to work with one pending queue and one
    completed queue. In the future, the number of queues can be varied for
    different throughput requirements at the cost of additional Memory and CPU
    consumption.
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, sid, rmgr, rts, zmq_info):

        if not isinstance(sid, str):
            raise EnTKTypeError(expected_type=str,
                            actual_type=type(sid))

        if not isinstance(rmgr, Base_ResourceManager):
            raise EnTKTypeError(expected_type=Base_ResourceManager,
                            actual_type=type(rmgr))

        if not isinstance(zmq_info, dict):
            raise EnTKTypeError(expected_type=dict,
                            actual_type=type(zmq_info))

        self._sid  = sid
        self._rmgr = rmgr
        self._rts  = rts

        # Utility parameters
        self._uid  = ru.generate_id('task_manager.%(counter)04d', ru.ID_CUSTOM)
        self._path = os.getcwd() + '/' + self._sid

        name = 'radical.entk.%s' % self._uid
        self._log  = ru.Logger  (name, path=self._path)
        self._prof = ru.Profiler(name, path=self._path)
        self._dh = ru.DebugHelper(name=name)

        self._tmgr_thread = None
        self._tmgr_terminate = None

        self._zmq_queue = None
        self._zmq_info  = zmq_info
        self._setup_zmq(zmq_info)


    # --------------------------------------------------------------------------
    #
    def _setup_zmq(self, zmq_info):

        self._prof.prof('zmq_setup_start', uid=self._uid)

        sid = self._sid
        self._zmq_queue = {
                'put' : ru.zmq.Putter(sid, url=zmq_info['put'], path=sid),
                'get' : ru.zmq.Getter(sid, url=zmq_info['get'], path=sid)}

        self._prof.prof('zmq_setup_stop', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def _tmgr(self, uid, rmgr, zmq_info):
        """
        **Purpose**: Method to be run by the tmgr thread. This method receives
                     a Task from the 'pending' queue and submits it to the RTS.
                     At all state transititons, they are synced (blocking) with
                     the AppManager in the master thread.

        **Details**: The AppManager can re-invoke the tmgr thread with this
                     function if the execution of the workflow is still
                     incomplete. There is also population of a dictionary,
                     placeholder_dict, which stores the path of each of the
                     tasks on the remote machine.
        """

        raise NotImplementedError('_tmgr() method not implemented in '
                                  'TaskManager for %s' % self._rts)


    # --------------------------------------------------------------------------
    #
    def _sync_with_master(self, obj, obj_type, qname):

        body = {'object': obj.as_dict(),
                'type'  : obj_type}

        if   obj_type == 'Task' : msg = obj.parent_stage['uid']
        elif obj_type == 'Stage': msg = obj.parent_pipeline['uid']
        else                    : msg = ''

        self._prof.prof('pub_sync', state=obj.state, uid=obj.uid, msg=msg)
        self._log.debug('%s (%s) to sync with amgr', obj.uid, obj.state)
        self._zmq_queue['put'].put(msgs=[body], qname=qname)


    # --------------------------------------------------------------------------
    #
    def _advance(self, obj, obj_type, new_state, qname):

        old_state = obj.state

        try:

            obj.state = new_state

            if   obj_type == 'Task' : msg = obj.parent_stage['uid']
            elif obj_type == 'Stage': msg = obj.parent_pipeline['uid']
            else                    : msg = None

            self._prof.prof('advance', uid=obj.uid, state=obj.state, msg=msg)
            self._log.info('Transition %s to %s', obj.uid, new_state)

            self._sync_with_master(obj, obj_type, qname)

        except Exception as ex:
            self._log.exception('Transition %s to state %s failed, error: %s',
                                obj.uid, new_state, ex)
            obj.state = old_state
            self._sync_with_master(obj, obj_type, qname)
            raise EnTKError(ex) from ex


    # --------------------------------------------------------------------------
    #
    def start_manager(self):
        """
        **Purpose**: Method to start the tmgr thread. The tmgr function
        is not to be accessed directly. The function is started in a separate
        thread using this method.
        """

        raise NotImplementedError('start_manager() method not implemented in '
                                  'TaskManager for %s' % self._rts)


    # --------------------------------------------------------------------------
    #
    def terminate_manager(self):
        """
        **Purpose**: Method to terminate the tmgr thread. This method is
        blocking as it waits for the tmgr thread to terminate (aka join).
        """

        try:
            if self._tmgr_thread:
                self._log.debug('Trying to terminate task manager.')
                if self._tmgr_terminate is not None:
                    if not self._tmgr_terminate.is_set():
                        self._tmgr_terminate.set()
                    self._log.debug('TMGR terminate is set %s',
                                    self._tmgr_terminate.is_set())
                if self.check_manager():
                    self._log.debug('TMGR thread is alive')
                    self._tmgr_thread.join(30)
                self._log.debug('TMGR thread joined')
                self._tmgr_thread = None

                self._log.info('Task manager thread closed')
                self._prof.prof('tmgr_term', uid=self._uid)

        except Exception as ex:
            self._log.exception('Could not terminate task manager thread')
            raise EnTKError(ex) from ex


    # --------------------------------------------------------------------------
    #
    def check_manager(self):
        """
        **Purpose**: Check if the tmgr thread is alive and running
        """

        # Return False if the thread does not exist
        if self._tmgr_thread:
            return self._tmgr_thread.is_alive()
        else:
            return False


# ------------------------------------------------------------------------------




# pylint: disable=protected-access

__copyright__ = 'Copyright 2014-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import warnings

from string import punctuation
from typing import Dict, List, Union

import radical.utils as ru

from .constants  import NAME_MESSAGE
from .exceptions import EnTKError, EnTKMissingError
from .exceptions import EnTKTypeError, EnTKValueError
from .states     import INITIAL, _task_state_values

VALID_TAGS = ['colocate', 'exclusive']


# ------------------------------------------------------------------------------
#
class ReqsMixin:

    # --------------------------------------------------------------------------
    #
    def _correct_deprecated_key(self, k):
        if k in self._deprecated_keys:
            warnings.warn('Key "%s" is obsolete, new key ' % k +
                          '"%s" will be used' % self._deprecated_keys[k],
                          DeprecationWarning, stacklevel=2)
            k = self._deprecated_keys[k]
        return k

    # --------------------------------------------------------------------------
    #
    def __setitem__(self, k, v):
        k = self._correct_deprecated_key(k)
        super().__setitem__(k, v)

    # --------------------------------------------------------------------------
    #
    def __setattr__(self, k, v):
        k = self._correct_deprecated_key(k)
        super().__setattr__(k, v)


# ------------------------------------------------------------------------------
#
class CpuReqs(ReqsMixin, ru.TypedDict):

    _check = True

    _schema = {
        'cpu_processes'   : int,
        'cpu_process_type': str,
        'cpu_threads'     : int,
        'cpu_thread_type' : str
    }

    _defaults = {
        'cpu_processes'   : 1,
        'cpu_process_type': None,
        'cpu_threads'     : 1,
        'cpu_thread_type' : None
    }

    # deprecated keys will cause a deprecation warning (ReqsMixin), and
    # correct keys, according to CpuReqs schema, will be provided
    _deprecated_keys = {
        'processes'          : 'cpu_processes',
        'process_type'       : 'cpu_process_type',
        'threads_per_process': 'cpu_threads',
        'thread_type'        : 'cpu_thread_type'
    }


# ------------------------------------------------------------------------------
#
class GpuReqs(ReqsMixin, ru.TypedDict):

    _check = True

    _schema = {
        'gpu_processes'   : float,
        'gpu_process_type': str,
        'gpu_threads'     : int,
        'gpu_thread_type' : str
    }

    _defaults = {
        'gpu_processes'   : 0.,
        'gpu_process_type': None,
        'gpu_threads'     : 1,
        'gpu_thread_type' : None
    }

    # deprecated keys will cause a deprecation warning (ReqsMixin), and
    # correct keys, according to GpuReqs schema, will be provided
    _deprecated_keys = {
        'processes'          : 'gpu_processes',
        'process_type'       : 'gpu_process_type',
        'threads_per_process': 'gpu_threads',
        'thread_type'        : 'gpu_thread_type'
    }


# ------------------------------------------------------------------------------
#
class Annotations(ru.TypedDict):

    _schema = {
        'inputs'    : [str],
        'outputs'   : [str],
        'depends_on': [str]
    }

    _defaults = {
        'inputs'    : [],
        'outputs'   : [],
        'depends_on': []
    }


# ------------------------------------------------------------------------------
#
class Task(ru.TypedDict):
    """
    A Task is an abstraction of a computational unit. In this case, a Task
    consists of its executable along with its required software environment,
    files to be staged as input and output.

    At the user level a Task is to be populated by assigning attributes.
    Internally, an empty Task is created and population using the `from_dict`
    function. This is to avoid creating Tasks with new `uid` as tasks with new
    `uid` offset the uid count file in radical.utils and can potentially affect
    the profiling if not taken care.

    .. data:: uid

        [type: `str` | default: `""`] A unique ID for the task. This attribute
        is optional, a unique ID will be assigned by RE if the field is not set,
        and cannot be re-assigned (immutable attribute).

    .. data:: name

        [type: `str` | default: `""`] A descriptive name for the task. This
        attribute can be used to map individual tasks back to application level
        workloads. Such characters as ',' or '_' are not allowed in a name.

    .. data:: state

        [type: `str` | default: `"DESCRIBED"`] Current state of the task, its
        initial value `re.states.INITIAL` is set during the task initialization,
        and this value is also appended into `state_history` attribute.

    .. data:: state_history

        [type: `list` | default: `["DESCRIBED"]`] List of the states obtained
        in temporal order. Every time new state is set, it is appended to
        `state_history` automatically.

    .. data:: executable

        [type: `str` | default: `""`] The executable to launch. The executable
        is expected to be either available via `$PATH` on the target resource,
        or to be an absolute path.

    .. data:: arguments

        [type: `list` | default: `[]`] List of arguments to be supplied to the
        `executable`.

    .. data:: environment

        [type: `dict` | default: `{}`] Environment variables to set in the
       environment before the execution process.

    .. data:: sandbox

        [type: `str` | default: `""`] This specifies the working directory of
        the task. By default, the task's uid is used as a sandbox for the task.

    .. data:: pre_launch

        [type: `list` | default: `[]`] List of actions (shell commands) to
        perform before the task is launched.

        Note that the set of shell commands given here are expected to load
        environments, check for work directories and data, etc. They are not
        expected to consume any significant amount of resources! Deviating from
        that rule will likely result in reduced overall throughput.

    .. data:: post_launch

        [type: `list` | default: `[]`] List of actions (shell commands) to
        perform after the task finishes.

        The same remarks as on `pre_launch` apply

    .. data:: pre_exec

        [type: `list` | default: `[]`] List of actions (shell commands) to
        perform after task is launched, but before rank(s) starts execution.

        The same remarks as on `pre_launch` apply

    .. data:: post_exec

        [type: `list` | default: `[]`] List of actions (shell commands) to
        perform right after rank(s) finishes.

        The same remarks as on `pre_launch` apply

    .. data:: cpu_reqs

        [type: `CpuReqs` | default: `CpuReqs()`] The CPU requirements for the
        current Task. The requirements are described in terms of the number of
        processes and threads to be run in this Task.

        The expected format is dict-like:

            task.cpu_reqs = {'cpu_processes'    : X,
                             'cpu_process_type' : None/'MPI',
                             'cpu_threads'      : Y,
                             'cpu_thread_type'  : None/'OpenMP'}

        This description means that the Task is going to spawn X processes and
        Y threads per each of these processes to run on CPUs. Hence, the total
        number of cpus required by the Task is `X * Y` for all the processes
        and threads to execute concurrently.

        By default, 1 CPU process and 1 CPU thread per process are requested.

    .. data:: gpu_reqs

        [type: `GpuReqs` | default: `GpuReqs()`] The GPU requirements for the
        current Task. The requirements are described in terms of the number of
        processes and threads to be run in this Task.

        The expected format is dict-like:

            task.gpu_reqs = {'gpu_processes'    : X,
                             'gpu_process_type' : None/'CUDA'/'ROCm',
                             'gpu_threads'      : Y,
                             'gpu_thread_type'  : None}

        This description means that each rank of the task is going to use X GPUs
        with Y GPU-threads.  By default, 0 GPUs are requested.

    .. data:: lfs_per_process

        [type: `int` | default: `0`] Local File Storage per process - amount of
        space (MB) required on the local file system of the node by the task.

    .. data:: mem_per_process

        [type: `int` | default: `0`] Amount of memory required by the task.

    .. data:: upload_input_data

        [type: `list` | default: `[]`] List of file names required to be
        transferred from a local client storage to the location of the task or
        data staging area before task starts.

    .. data:: copy_input_data

        [type: `list` | default: `[]`] List of file names required to be copied
        from another task-/pilot-sandbox to a current task (or data staging
        area) before task starts.

        Example of a file location format:

            $Pipeline_%s_Stage_%s_Task_%s, where %s is replaced entity name/uid

            # output.txt is copied from a t1 task to a current task sandbox
            t2.copy_input_data = ['$Pipeline_%s_Stage_%s_Task_%s/output.txt' %
                                  (p.name, s1.name, t1.name)]

    .. data:: link_input_data

        [type: `list` | default: `[]`] List of file names required to be
        symlinked from another task-/pilot-sandbox to a current task (or data
        staging area) before task starts.

    .. data:: move_input_data

        [type: `list` | default: `[]`] List of file names required to be moved
        from another task-/pilot-sandbox to a current task (or data staging
        area) before task starts.

    .. data:: copy_output_data

        [type: `list` | default: `[]`] List of file names required to be copied
        from a current task to another task-/pilot-sandbox (or data staging
        area) when a task is finished.

        Example of defining data to be copied:

            # results.txt will be copied to a data staging area `$SHARED`
            t.copy_output_data = ['results.txt > $SHARED/results.txt']

    .. data:: link_output_data

        [type: `list` | default: `[]`] List of file names required to be
        symlinked from a current task to another task-/pilot-sandbox (or data
        staging area) when a task is finished.

    .. data:: move_output_data

        [type: `list` | default: `[]`] List of file names required to be moved
        from a current task to another task-/pilot-sandbox (or data staging
        area) when a task is finished.

    .. data:: download_output_data

        [type: `list` | default: `[]`] List of file names required to be
        downloaded from a current task to a local client storage space when a
        task is finished.

        Example of defining data to be downloaded:

            # results.txt is transferred to a local client storage space
            t.download_output_data = ['results.txt']

    .. data:: stdout

        [type: `str` | default: `""`] The name of the file to store stdout. If
        not set then the name of the following format will be used: `<uid>.out`.

    .. data:: stderr

        [type: `str` | default: `""`] The name of the file to store stderr. If
        not set then the name of the following format will be used: `<uid>.err`.

    .. data:: stage_on_error

        [type: `bool` | default: `False`] Flag to allow staging out data if
        task got failed (output staging is normally skipped on `FAILED` or
        `CANCELED` tasks).

    .. data:: exit_code

        [type: `int` | default: `None`] Get the exit code for finished tasks:
        0 - for successful tasks; 1 - for failed tasks.

    .. data:: exception

        [type: `str` | default: `None`] Get the representation of the exception
        which caused the task to fail.

    .. data:: exception_detail

        [type: `str` | default: `None`] Get additional details (traceback or
        error messages) to the exception which caused this task to fail.

    .. data:: path

        [type: `str` | default: `""`] Get the path of the task on the remote
        machine. Useful to reference files generated in the current task.

    .. data:: tags

        [type: `dict` | default: `None`] The tags for the task that can be
        used while scheduling by the RTS (configuration specific tags, which
        influence task scheduling and execution, e.g., tasks co-location).

    .. data:: rts_uid

        [type: `str` | default: `None`] Unique RTS ID of the current task.

    .. data:: parent_stage

        [type: `dict` | default: `{'uid': None, 'name': None}`] Identification
        of the stage, which contains the current task.

    .. data:: parent_pipeline

        [type: `dict` | default: `{'uid': None, 'name': None}`] Identification
        of the pipeline, which contains the current task.

    .. data:: annotations

        [type: `Annotations` | default: `None`] Annotations to describe task's
        input and output files, and sets dependencies between tasks.

    Read-only attributes
    --------------------

    .. property:: luid

        [type: `str`] Unique ID of the current task (fully qualified).

            > task.luid
            pipe.0001.stage.0004.task.0234
    """

    _check = True
    _uids  = []

    _schema = {
        'uid'                  : str,
        'name'                 : str,
        'state'                : str,
        'state_history'        : [str],
        'executable'           : str,
        'arguments'            : [str],
        'environment'          : {str: str},
        'sandbox'              : str,
        'pre_launch'           : [str],
        'post_launch'          : [str],
        'pre_exec'             : [str],
        'post_exec'            : [str],
        'cpu_reqs'             : CpuReqs,
        'gpu_reqs'             : GpuReqs,
        'lfs_per_process'      : int,
        'mem_per_process'      : int,
        'upload_input_data'    : [str],
        'copy_input_data'      : [str],
        'link_input_data'      : [str],
        'move_input_data'      : [str],
        'copy_output_data'     : [str],
        'link_output_data'     : [str],
        'move_output_data'     : [str],
        'download_output_data' : [str],
        'stdout'               : str,
        'stderr'               : str,
        'stage_on_error'       : bool,
        'exit_code'            : int,
        'exception'            : str,
        'exception_detail'     : str,
        'path'                 : str,
        'tags'                 : {str: None},
        'rts_uid'              : str,
        'parent_stage'         : {str: None},
        'parent_pipeline'      : {str: None},
        'annotations'          : Annotations
    }

    # guaranteed attributes with default non-initialized values
    _defaults = {
        'uid'                  : '',
        'name'                 : '',
        'state'                : '',
        'state_history'        : [],
        'executable'           : '',
        'arguments'            : [],
        'environment'          : {},
        'sandbox'              : '',
        'pre_launch'           : [],
        'post_launch'          : [],
        'pre_exec'             : [],
        'post_exec'            : [],
        'cpu_reqs'             : CpuReqs(),
        'gpu_reqs'             : GpuReqs(),
        'lfs_per_process'      : 0,
        'mem_per_process'      : 0,
        'upload_input_data'    : [],
        'copy_input_data'      : [],
        'link_input_data'      : [],
        'move_input_data'      : [],
        'copy_output_data'     : [],
        'link_output_data'     : [],
        'move_output_data'     : [],
        'download_output_data' : [],
        'stdout'               : '',
        'stderr'               : '',
        'stage_on_error'       : False,
        'exit_code'            : None,
        'exception'            : None,
        'exception_detail'     : None,
        'path'                 : '',
        'tags'                 : None,
        'rts_uid'              : None,
        'parent_stage'         : {'uid': None, 'name': None},
        'parent_pipeline'      : {'uid': None, 'name': None}
    }

    # --------------------------------------------------------------------------
    #
    def __init__(self, from_dict=None):

        from_dict = from_dict or {}
        if not isinstance(from_dict, dict):
            raise EnTKTypeError(expected_type=dict, actual_type=type(from_dict))

        super().__init__()

        attrs = list(from_dict.keys())
        # `uid` is allowed to be set during initialization only
        if not from_dict.get('uid'):
            uid = ru.generate_id('task.%(counter)06d', ru.ID_CUSTOM)
        else:
            attrs.remove('uid')
            uid = super()._verify_setter('uid', from_dict['uid'])
            if any(s in uid for s in punctuation.replace('.', '')):
                raise EnTKError('Incorrect symbol for attribute "uid" (%s). %s'
                                % (uid, NAME_MESSAGE))

        self._data['uid'] = uid

        # ensure that "state" and "state_history" are handled correctly
        self['state'] = from_dict.get('state') or INITIAL
        if 'state' in from_dict:
            attrs.remove('state')
        if 'state_history' in from_dict:
            self['state_history'] = from_dict['state_history']
            attrs.remove('state_history')

        for attr in attrs:
            self[attr] = from_dict[attr]

    # --------------------------------------------------------------------------
    #
    def __eq__(self, other):

        if not isinstance(other, Task):
            return False

        return self['uid'] == other['uid']

    # --------------------------------------------------------------------------
    #
    def __ne__(self, other):

        return not self == other

    # --------------------------------------------------------------------------
    #
    def __hash__(self):

        return hash(self['uid'])

    # --------------------------------------------------------------------------
    #
    def _post_verifier(self, k, v):

        if not v:
            return

        if k == 'uid':
            raise EnTKError('Task.uid is not allowed to be re-assigned')

        elif k == 'name':
            if any(symbol in v for symbol in punctuation.replace('.', '')):
                raise EnTKError('Incorrect symbol for attribute "%s" (%s). %s'
                                % (k, v, NAME_MESSAGE))

        elif k == 'state':
            if v not in _task_state_values:
                raise EnTKValueError(obj=self['uid'], attribute=k,
                                    expected_value=list(_task_state_values),
                                    actual_value=v)
            self['state_history'].append(v)

        elif k == 'state_history':
            for _v in v:
                if _v not in _task_state_values:
                    raise EnTKValueError(obj=self['uid'],
                                    attribute='state_history element',
                                    expected_value=list(_task_state_values),
                                    actual_value=_v)

        elif k == 'tags':
            if any(tag not in VALID_TAGS for tag in v):
                raise EnTKError(
                    'Incorrect structure for attribute "%s" of object %s' %
                    (k, self['uid']))

    # --------------------------------------------------------------------------
    #
    def _verify_setter(self, k, v):
        v = super()._verify_setter(k, v)
        self._post_verifier(k, v)
        return v

    # --------------------------------------------------------------------------
    #
    def _verify(self):
        verify_attrs = list(self.keys())
        verify_attrs.remove('uid')

        for k in verify_attrs:
            self._post_verifier(k, self[k])

    # --------------------------------------------------------------------------
    #
    @property
    def luid(self):
        """
        Unique ID of the current task (fully qualified).

        Example:
            > task.luid
            pipe.0001.stage.0004.task.123456

        :luid: Returns the fully qualified uid of the current task
        :type: str
        """

        # TODO: cache

        p_elem = self['parent_pipeline'].get('name')
        s_elem = self['parent_stage'].get('name')
        t_elem = self['name']

        if not p_elem: p_elem = self['parent_pipeline']['uid']
        if not s_elem: s_elem = self['parent_stage']['uid']
        if not t_elem: t_elem = self['uid']

        return '%s.%s.%s' % (p_elem, s_elem, t_elem)

    # --------------------------------------------------------------------------
    #
    def annotate(self, inputs: Union[Dict, List, str, None] = None,
                       outputs: Union[List, str, None] = None) -> None:
        """
        Adds dataflow annotations with provided input and output files,
        and defines dependencies between tasks.

        Attributes:
            inputs (list, optional): List of input files. If a file is
                produced by the previously executed task, then the
                corresponding input element is provided as a dictionary with
                the task instance as a key.
                Example: inputs=['file1', {task0: 'file2', task1: ['file2']}]

            outputs (list, optional): List of produced/generated files.
        """
        ta = self.setdefault('annotations', Annotations())

        # TODO: have a unified file representation (file name within task
        #       sandbox or file full path, considering files outside of the
        #       task sandbox) - set an expand procedure. Ensure that it is
        #       the same for inputs and outputs

        if inputs:
            for i in ru.as_list(inputs):

                if isinstance(i, str):
                    ta.inputs.append(i)

                elif isinstance(i, dict):
                    for task, task_files in i.items():
                        if task.uid not in ta.depends_on:
                            ta.depends_on.append(task.uid)
                        for tf in ru.as_list(task_files):
                            # corresponding validation of dependencies
                            # is part of the Pipeline validation process
                            ta.inputs.append('%s:%s' % (task.uid, tf))

        if outputs:
            ta.outputs.extend(ru.as_list(outputs))
            if len(ta.outputs) != len(set(ta.outputs)):
                warnings.warn('Annotated outputs for %s ' % self['uid'] +
                              'includes duplication(s)')

    # --------------------------------------------------------------------------
    #
    def from_dict(self, d):
        """Re-initialization, resets all attributes with provided input data."""
        self.__init__(from_dict=d)

    # --------------------------------------------------------------------------
    #
    def _validate(self):
        """
        Purpose: Validate that the state of the task is 'DESCRIBED' and that an
        executable has been specified for the task.
        """

        if self['uid'] in Task._uids:
            raise EnTKError('Task ID %s already exists' % self['uid'])
        else:
            Task._uids.append(self['uid'])

        if self['state'] is not INITIAL:
            raise EnTKValueError(obj=self['uid'], attribute='state',
                                 expected_value=INITIAL,
                                 actual_value=self['state'])

        if not self['executable']:
            raise EnTKMissingError(obj=self['uid'],
                                   missing_attribute='executable')

# ------------------------------------------------------------------------------




__copyright__ = 'Copyright 2020-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'


NAME_MESSAGE = "Valid object names can contain letters, numbers and '.'. " \
               "Any other character will not be allowed in versions >= 1.5.10"



# pylint: disable=protected-access

__copyright__ = 'Copyright 2017-2018, http://radical.rutgers.edu'
__author__    = 'Vivek Balasubramanian <vivek.balasubramanian@rutgers.edu>'
__license__   = 'MIT'


import os
import queue
import pickle

import threading       as mt

import radical.pilot   as rp

from ...exceptions       import EnTKError
from ...                 import states, Task
from ..base              import Base_TaskManager
from .task_processor     import create_td_from_task, create_task_from_rp


# ------------------------------------------------------------------------------
#
class TaskManager(Base_TaskManager):
    '''
    A Task Manager takes the responsibility of dispatching tasks it receives
    from a queue for execution on to the available resources using a runtime
    system. In this case, the runtime system being used RADICAL Pilot. Once
    the tasks have completed execution, they are pushed on to another queue for
    other components of EnTK to access.


    :arguments:
        :rmgr:              (ResourceManager) Object to be used to access the
                            Pilot where the tasks can be submitted

    Currently, EnTK is configured to work with one pending queue and one
    completed queue. In the future, the number of queues can be varied for
    different throughput requirements at the cost of additional Memory and CPU
    consumption.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, sid, rmgr, zmq_info):

        super(TaskManager, self).__init__(sid, rmgr, rts='radical.pilot',
                                          zmq_info=zmq_info)
        self._rts_runner = None
        self._zmq_info   = zmq_info

        self._log.info('Created task manager object: %s', self._uid)
        self._prof.prof('tmgr_create', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def _tmgr(self, uid, rmgr, zmq_info):
        '''
        **Purpose**: This method has 2 purposes: receive tasks from the
                     'pending' queue, start a new thread that processes these
                     tasks and submits to the RTS.

                     The new thread is responsible for pushing completed tasks
                     (returned by the RTS) to the dequeueing queue. It also
                     converts Tasks into TDs and CUs into (partially described)
                     Tasks. This conversion is necessary since the current RTS
                     is RADICAL Pilot. Once Tasks are recovered from a CU, they
                     are then pushed to the completed queue. At all state
                     transitions, they are synced (blocking) with the
                     AppManager in the master process.

        **Details**: The AppManager can re-invoke the tmgr thread with this
                     function if the execution of the workflow is still
                     incomplete. There is also population of a dictionary,
                     `placeholders`, which stores the path of each of the
                     tasks on the remote machine.
        '''

        session = rp.Session(uid=self._sid,
                             _reg_addr=rmgr.pmgr.cfg.reg_addr,
                             _role=rp.Session._DEFAULT)
        # FIXME: use a non-primary session for rp.TaskManager
        session._role = rp.Session._PRIMARY
        tmgr = rp.TaskManager(session=session)

        self._submitted_tasks = dict()
        self._total_res       = {'cores': 0, 'gpus': 0}

        try:

            self._setup_zmq(zmq_info)

            self._prof.prof('tmgr thread started', uid=self._uid)
            self._log.info('Task Manager thread started')

            # Queue for communication between threads of this thread
            task_queue     = queue.Queue()
            tasks_per_item = 1000  # sets a size of item for `Queue.put`

            # Pickle file for task id history.
            # TODO: How do you take care the first execution.
            pkl_path = self._path + '/.task_submitted.pkl'
            if os.path.exists(pkl_path):
                with open(pkl_path, 'rb') as f:
                    self._submitted_tasks = pickle.load(f)

            # Start a thread to receive tasks and push to RTS
            self._rts_runner = mt.Thread(target=self._process_tasks,
                                         args=(task_queue, tmgr))
            self._rts_runner.daemon = True
            self._rts_runner.start()

            self._prof.prof('tmgr infrastructure setup done', uid=uid)

            # While we are supposed to run and the thread that does the work is
            # alive go.
            while not self._tmgr_terminate.is_set() and \
                      self._rts_runner.is_alive():

                try:

                    msgs = self._zmq_queue['get'].get_nowait(
                            qname='pending', timeout=100)

                    if msgs:

                        for msg in msgs:

                            if msg['type'] == 'workload':

                                tasks = msg['body']
                                self._log.debug('Task queue: ready workload '
                                                'with %s task(s)', len(tasks))

                                queue_items = [
                                    tasks[i:i + tasks_per_item] for i in
                                    range(0, len(tasks), tasks_per_item)]

                                for item in queue_items:
                                    task_queue.put(item)

                            elif msg['type'] == 'rts':
                                self._update_resource(msg['body'], tmgr)

                            else:
                                self._log.error('TMGR receiver wrong message type')

                except Exception as e:
                    self._log.exception('Error in task execution')
                    raise EnTKError(e) from e

            self._log.debug('Exited TMGR main loop')

        except KeyboardInterrupt:

            self._log.exception('Execution interrupted (probably by Ctrl+C), '
                                'cancel tmgr thread gracefully...')
            raise


        except Exception as e:

            self._log.exception('task %s failed')
            raise EnTKError(e) from e

        finally:

            self._prof.prof('tmgr_term', uid=uid)

            if self._rts_runner:
                self._rts_runner.join()
            tmgr.close()

            self._log.debug('TMGR RTS Runner joined')

            self._prof.close()
            self._log.debug('TMGR profile closed')


    # --------------------------------------------------------------------------
    #
    def _update_resource(self, pilot, tmgr):
        '''
        Update used pilot.
        '''

        # Busy wait unit RP TMGR exists. Does some other way make sense?
        self._log.debug('Adding pilot.')

        curr_pilot = tmgr.list_pilots()
        if curr_pilot:
            self._log.debug('Got old pilots')
            tmgr.remove_pilots(pilot_ids=curr_pilot)
        tmgr.add_pilots(pilot)

        self._total_res = {'cores': pilot['description']['cores'],
                           'gpus' : pilot['description']['gpus']}
        self._log.debug('Added new pilot')


    # --------------------------------------------------------------------------
    #
    def _process_tasks(self, task_queue, tmgr):
        '''
        **Purpose**: The new thread that gets spawned by the main tmgr thread
                     invokes this function. This function receives tasks from
                     'task_queue' and submits them to the RADICAL Pilot RTS.
        '''

        placeholders = dict()
        placeholders['__by_name__'] = dict()
        placeholders_by_name = placeholders['__by_name__']
        placeholder_lock = mt.Lock()

        # ----------------------------------------------------------------------
        def load_placeholder(task):
            with placeholder_lock:
                parent_pipeline = str(task.parent_pipeline['uid'])
                parent_stage = str(task.parent_stage['uid'])

                if parent_pipeline not in placeholders:
                    placeholders[parent_pipeline] = dict()

                if parent_stage not in placeholders[parent_pipeline]:
                    placeholders[parent_pipeline][parent_stage] = dict()

                if None not in [parent_pipeline, parent_stage, task.uid]:
                    placeholders[parent_pipeline][parent_stage][task.uid] = \
                                                            {'path': task.path,
                                                             'uid': task.uid}

                parent_pipeline_n = str(task.parent_pipeline['name'])
                parent_stage_n = str(task.parent_stage['name'])

                if parent_pipeline_n not in placeholders_by_name:
                    placeholders_by_name[parent_pipeline_n] = dict()

                if parent_stage_n not in placeholders_by_name[parent_pipeline_n]:
                    placeholders_by_name[parent_pipeline_n][parent_stage_n] = dict()

                if None not in [parent_pipeline_n, parent_stage_n, task.name]:
                    placeholders_by_name[parent_pipeline_n][parent_stage_n][task.name] = \
                                                            {'path': task.path,
                                                             'uid': task.uid}

        # ----------------------------------------------------------------------
        def task_state_cb(rp_task, state):

            try:
                self._log.debug('Task %s in state %s', rp_task.uid,
                                                       rp_task.state)

                if rp_task.state in rp.FINAL:

                    task = create_task_from_rp(rp_task, self._log, self._prof)

                    self._advance(task, 'Task', states.COMPLETED, 'cb-to-sync')

                    load_placeholder(task)

                    tdict = task.as_dict()

                    self._zmq_queue['put'].put(qname='completed', msgs=[tdict])
                    self._log.info('Pushed task %s with state %s to completed',
                                   task.uid, task.state)

            except KeyboardInterrupt as ex:
                self._log.exception('Execution interrupted (probably by Ctrl+C)'
                                    ' exit callback thread gracefully...')
                raise KeyboardInterrupt(ex) from ex

            except Exception as ex:
                self._log.exception('Error in RP callback thread: %s', ex)
                raise EnTKError(ex) from ex
        # ----------------------------------------------------------------------

        tmgr.register_callback(task_state_cb)

        try:
            pkl_path = self._path + '/.task_submitted.pkl'

            while not self._tmgr_terminate.is_set():

                try:
                    body = task_queue.get(block=True, timeout=10)
                except queue.Empty:
                    # Ignore, we don't always have new tasks to run
                    continue

                task_queue.task_done()
                self._log.debug('Task queue: got workload with '
                                '%s task(s)', len(body or []))

                if not body:
                    continue

                bulk_tasks = list()

                for msg in body:

                    task = Task(from_dict=msg)
                    load_placeholder(task)
                    bulk_tasks.append(create_td_from_task(
                                        task, placeholders,
                                        self._submitted_tasks, pkl_path,
                                        self._sid, self._log, self._prof))

                    self._advance(task, 'Task', states.SUBMITTING,
                                      'tmgr-to-sync')

                if bulk_tasks:
                    tmgr.submit_tasks(bulk_tasks)

            self._log.debug('Exited RTS main loop. TMGR terminating')
        except KeyboardInterrupt:
            self._log.exception('Execution interrupted (probably by Ctrl+C), '
                                'cancel task processor gracefully...')
            raise

        except Exception as e:
            self._log.exception('%s failed', self._uid)
            raise EnTKError(e) from e


    # --------------------------------------------------------------------------
    #
    def start_manager(self):
        '''
        **Purpose**: Method to start the tmgr thread. The tmgr function
                     is not to be accessed directly. The function is started
                     in a separate thread using this method.
        '''
        # pylint: disable=attribute-defined-outside-init, access-member-before-definition
        if self._tmgr_thread:
            self._log.warn('tmgr thread already running!')
            return

        try:
            self._prof.prof('creating tmgr thread', uid=self._uid)
            self._tmgr_terminate = mt.Event()

            # preserve session before forking
            session = self._rmgr.session
            self._rmgr._session = None

            self._tmgr_thread = mt.Thread(target=self._tmgr,
                                          name='task-manager',
                                          args=(self._uid,
                                                self._rmgr,
                                                self._zmq_info))
            self._tmgr_thread.daemon = True

            self._rmgr._session = session

            self._log.info('Starting task manager thread')
            self._prof.prof('starting tmgr thread', uid=self._uid)

            self._tmgr_thread.start()
            self._log.debug('tmgr thread "task-manager" started')

            return True

        except Exception as e:

            self._log.exception('Task manager not started')
            self.terminate_manager()
            raise EnTKError(e) from e


# ------------------------------------------------------------------------------
# pylint: enable=attribute-defined-outside-init, access-member-before-definition



#!/usr/bin/env python3

import functools

import radical.entk as re

#
# Similar to the first tutorial exercise, we will implement an ensemble of
# simulation pipelines where each pipeline prepares data, runs a set of
# simulation on the data, and then accumulates results.  Other than before
# though we will add dynamicity to the application:
#
#   - after each simulation stage, evaluate intermediate data
#   - if data diverge (sum is larger than `LIMIT`)
#     - re-seed the pipeline with new data
#   - else
#     - run another simulation steps
#   - finish simulation after at most `MAX_ITER` steps and then collect results
#
# This exercise will thus show how the workflow can be adapted at runtime, by
# inserting new pipeline stages or stopping the execution of individual
# pipelines.
#

N_PIPELINES   =    5                 # number of simulation pipeline
N_SIMULATIONS =    5                 # number of simulations per pipeline
MAX_ITER      =   10                 # max number of simulation steps
LIMIT         =  100 * 1000 * 1000   # max intermediate result


# we want to change pipelines on the fly, thus want to keep track
# of all pipelines.  We identify pipelines by an application specified
# name `pname`
ensemble = dict()


# ------------------------------------------------------------------------------
#
def get_stage_1(sandbox, pname):
    '''
    Create a stage which seeds (or re-seeds) a simulation pipeline with a random
    integer as input data.

    The returned stage will include a `post_exec` callback which outputs the new
    data seed after completion of the stage
    '''

    t1 = re.Task()
    t1.executable = '/bin/sh'
    t1.arguments  = ['-c', 'od -An -N1 -i /dev/random']
    t1.stdout     = 'random.txt'
    t1.sandbox    = sandbox
    t1.download_output_data = ['random.txt > %s.random.txt' % pname]

    s1 = re.Stage()
    s1.add_tasks(t1)

    # --------------------------------------------------------------------------
    # use a callback after that stage completed for output of the seed value
    def post_exec(_stage, _pname):
        fname = '%s.random.txt' % _pname
        seed  = int(open(fname, encoding='utf-8').read().split()[-1])
        print(_pname, 'rand  --- - %10d' % seed)
    # --------------------------------------------------------------------------
    s1.post_exec = functools.partial(post_exec, s1, pname)

    return s1


# ------------------------------------------------------------------------------
#
def get_stage_2(sandbox, pname, iteration=0):
    '''
    The second pipeline stage is again the simulation stage: it consists of
    `N_SIMULATIONS` tasks which compute the n'th power of the input data (last
    line of `random.txt`).  Another `post_exec` callback will, after all tasks
    are done, evaluate the intermediate data and decide how to continue:

        - if result    > LIMIT   : new seed  (add stages 1 and 2)
        - if iteration > MAX_ITER: abort     (add stage 3)
        - else                   : continue  (add stage 2 again)

    For simplicity we keep track of iterations within the pipeline instances
    (`pipeline.iteration`).  A 'proper' implementation may want to create
    a subclass from `re.Pipeline` which hosts that counter as `self._iteration`.
    '''

    pipeline = ensemble[pname]

    # this is the pseudo simulation we iterate on the second stage
    sim = 'echo "($(tail -qn 1 random.txt) + %(iteration)d) ^ %(ensemble_id)d"'\
          '| bc'

    # second stage: create N_SIMULATIONS tasks to compute the n'th power
    s2 = re.Stage()
    pipeline.iteration = iteration                                # type: ignore
    for i in range(N_SIMULATIONS):
        t2 = re.Task()
        t2.executable = '/bin/sh'
        t2.arguments  = ['-c', sim % {'iteration'  : iteration,
                                      'ensemble_id': i}]
        t2.stdout     = 'power.%03d.txt' % i
        t2.sandbox    = sandbox
        t2.download_output_data = ['%s > %s.%s' % (t2.stdout, pname, t2.stdout)]
        s2.add_tasks(t2)

    # --------------------------------------------------------------------------
    # add a callback after that stage's completed which checks the
    # intermediate results:
    def post_exec(stage, pname):

        pipeline = ensemble[pname]
        iteration = pipeline.iteration

        # continue to iterate - check intermediate data
        result = 0
        for task in stage.tasks:
            fname   = '%s.%s' % (pname, task.stdout)
            data    = open(fname, encoding='utf-8').read()
            result += int(data.split()[-1])

        if result > LIMIT:
            # simulation diverged = reseed the pipeline (add new stages 1 and 2)
            print(pipeline.name, 'seed  %3d - %10d' % (iteration, result))
            pipeline.add_stages(get_stage_1(sandbox, pname))
            pipeline.add_stages(get_stage_2(sandbox, pname))

        elif iteration > MAX_ITER:
            # iteration limit reached, discontinue pipeline (add final stage 3)
            print(pipeline.name, 'break %3d - %10d' % (iteration, result))
            pipeline.add_stages(get_stage_3(sandbox, pname))

        else:
            # continue to iterate (increase the iteration counter)
            print(pipeline.name, 'iter  %3d - %10d' % (iteration, result))
            pipeline.add_stages(get_stage_2(sandbox, pname, iteration + 1))

    # --------------------------------------------------------------------------
    s2.post_exec = functools.partial(post_exec, s2, pname)

    return s2


# ------------------------------------------------------------------------------
#
def get_stage_3(sandbox, pname):
    '''
    This pipeline has reached its iteration limit without exceeding the
    simulation results diverging beyond `LIMIT`.  This final stage will collect
    the simulation results and report the final data, again via an `post-exec`
    hook.
    '''

    # third stage: compute sum over all powers
    t3 = re.Task()
    t3.executable = '/bin/sh'
    t3.arguments  = ['-c', 'tail -qn 1 power.*.txt | paste -sd+ | bc']
    t3.stdout     = 'sum.txt'
    t3.sandbox    = sandbox
    t3.download_output_data = ['%s > %s.%s' % (t3.stdout, pname, t3.stdout)]

    # download the result while renaming to get unique files per pipeline
    t3.download_output_data = ['sum.txt > %s.sum.txt' % pname]

    s3 = re.Stage()
    s3.add_tasks(t3)

    # --------------------------------------------------------------------------
    # use a callback after that stage completed for output of the final result
    def post_exec(_, pname):
        fname  = '%s.sum.txt' % pname
        result = int(open(fname, encoding='utf-8').read())
        print(pname, 'final %3d - %10d' % (MAX_ITER, result))
    # --------------------------------------------------------------------------
    s3.post_exec = functools.partial(post_exec, s3, pname)

    return s3


# ------------------------------------------------------------------------------
#
def generate_pipeline(pname):
    '''
    We generate essentially the same pipeline as in `radical_entk_1.py` with
    three stages:

      1) generate a random seed as input data
      2) evolve a model based on that input data via a set of ensembles
      3) derive a common metric across the model results

    However, we will iterate the model (stage 2) multiple times and check for
    intermediate results.  Further, we will cancel all pipelines whose
    intermediate result is larger than some threshold and will instead replace
    that pipeline with a newly seeded pipeline.  We break after a certain number
    of iterations and expect the result to be biased toward smaller seeds.
    '''

    # create and register pipeline
    p = re.Pipeline()
    p.name = pname
    ensemble[pname] = p

    # all tasks in this pipeline share the same sandbox
    sandbox = pname

    # first stage: create 1 task to generate a random seed number
    s1 = get_stage_1(sandbox, pname)

    # second stage: create N_SIMULATIONS tasks to compute the n'th power
    # of that number (this stage runs at least once)
    s2 = get_stage_2(sandbox, pname)

    # the third stage is added dynamically after convergence, so we return
    # the pipeline with the initial two stages
    p.add_stages(s1)
    p.add_stages(s2)

    return p


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # create application manager which executes our ensemble
    appman = re.AppManager()

    # assign resource request description to the application manager using
    # three mandatory keys: target resource, walltime, and number of cpus
    appman.resource_desc = {
        'resource': 'local.localhost',
        'walltime': 10,
        'cpus'    : 1
    }

    # create an ensemble of n simulation pipelines
    for cnt in range(N_PIPELINES):
        generate_pipeline(pname='pipe.%03d' % cnt)

    # assign the workflow to the application manager, then
    # run the ensemble and wait for completion
    appman.workflow = set(ensemble.values())
    appman.run()

    # check results which were staged back
    for cnt in range(N_PIPELINES):
        pipe_fname  = 'pipe.%03d.sum.txt' % cnt
        pipe_result = int(open(pipe_fname, encoding='utf-8').read())
        print('%18d - %10d' % (cnt, pipe_result))


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import sys
import time

from radical.entk import Pipeline, Stage, Task, AppManager

n_pipes  = int(sys.argv[1])
n_stages = int(sys.argv[2])
n_tasks  = int(sys.argv[3])


# ------------------------------------------------------------------------------
#
def main():

    start = time.time()


    pipelines = set()
    for _ in range(n_pipes):

        p = Pipeline()

        for _ in range(n_stages):

            s = Stage()

            for _ in range(n_tasks):

                t = Task()
                t.executable = '/bin/true'

                s.add_tasks(t)

            p.add_stages(s)

        pipelines.add(p)


    appman = AppManager(autoterminate=True, rts='mock')
    appman.resource_desc = {'resource': 'local.localhost_test',
                            'cpus'    :  8,
                            'walltime': 10}
    appman.workflow      = pipelines

    appman.run()
    appman.terminate()

    print('%10.1f' % (time.time() - start))


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

  # import yappi
  # yappi.start(builtins=False)

    main()

  # yappi.get_thread_stats().print_all()
  # stats = yappi.convert2pstats(yappi.get_func_stats())
  # stats.dump_stats('pstats.prof')
  #
  # stats.print_stats()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import os
import pprint
import radical.utils as ru
import radical.entk as re
import radical.analytics as ra

__copyright__ = 'Copyright 2013-2018, http://radical.rutgers.edu'
__license__ = 'MIT'

"""
This example illustrates the use of the method ra.Session.get().
Modified from examples under RADICAL Analytics
"""

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    loc = './re.session.two.vivek.017759.0012'
    src = os.path.dirname(loc)
    sid = os.path.basename(loc)
    session = ra.Session(src=src, sid=sid, stype='radical.entk')

    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print(separator + message + separator)

    # and here we go. As seen in example 01, we use ra.Session.list() to get the
    # name of all the types of entity of the session.
    etypes = session.list('etype')
    pprint.pprint(etypes)

    # We limit ourselves to the type 'task'. We use the method
    # ra.Session.get() to get all the objects in our session with etype 'task':
    ppheader("properties of the entities with etype 'task'")
    tasks = session.get(etype='task')
    pprint.pprint(tasks)


    # Mmmm, still a bit too many entities. We limit our analysis to a single
    # task. We use ra.Session.get() to select all the objects in the
    # session with etype 'task' and uid 'task.0000' and return them into a
    # list:
    ppheader("properties of the entities with etype 'task' and uid 'task.0000'")
    task = session.get(etype='task', uid='task.0000')
    pprint.pprint(task)


    # We may want also to look into the states of this task:
    ppheader("states of the entities with uid 'task.0000'")
    states = task[0].states
    pprint.pprint(states)

    # and extract the state we need. For example, the state 'SCHEDULED', that
    # indicates that the task has been scheduled. To refer to the state 'SCHEDULED',
    # and to all the other states of RADICAL-Pilot, we use the re.states.SCHEDULED property
    # that guarantees type checking.
    ppheader("Properties of the state re.SCHEDULED of the entities with uid 'task.0000'")
    state = task[0].states[re.states.SCHEDULED]
    pprint.pprint(state)

    # Finally, we extract a property we need from this state. For example, the
    # timestamp of when the task has been created, i.e., the property 'time' of
    # the state SCHEDULED:
    ppheader("Property 'time' of the state re.states.SCHEDULED of the entities with uid 'task.000000'")
    timestamp = task[0].states[re.states.SCHEDULED][ru.TIME]
    pprint.pprint(timestamp)

    # ra.Session.get() can also been used to to get all the entities in our
    # session that have a specific state. For example, the following gets all
    # the types of entity that have the state 'SCHEDULED':
    ppheader("Entities with state re.states.SCHEDULED")
    entities = session.get(state=re.states.SCHEDULED)
    pprint.pprint(entities)

    # We can then print the timestamp of the state 'SCHEDULED' for all the entities
    # having that state by using something like:
    ppheader("Timestamp of all the entities with state re.states.SCHEDULED")
    timestamps = [entity.states[re.states.SCHEDULED][ru.TIME] for entity in entities]
    pprint.pprint(timestamps)

    # We can also create tailored data structures for our analyis. For
    # example, using tuples to name entities, state, and timestamp:
    ppheader("Named entities with state re.states.SCHEDULED and its timestamp")
    named_timestamps = [(entity.uid,
                        entity.states[re.states.SCHEDULED][ru.STATE],
                        entity.states[re.states.SCHEDULED][ru.TIME]) for entity in entities]
    pprint.pprint(named_timestamps)




__copyright__ = "Copyright 2017-2019, http://radical.rutgers.edu"
__author__    = "RADICAL Team <radical@rutgers.edu>"
__license__   = "MIT"


import os
import threading     as mt

import radical.utils as ru

from ..states     import INITIAL, FINAL
from ..states     import SUSPENDED, SCHEDULED, SCHEDULING, DONE, FAILED
from ..task       import Task
from ..exceptions import EnTKError


# ------------------------------------------------------------------------------
#
class WFprocessor(object):
    """
    An WFprocessor (workflow processor) takes the responsibility of dispatching
    tasks from the various pipelines of the workflow according to their relative
    order to the TaskManager. All state updates are relflected in the AppManager
    as we operate on the reference of the same workflow object. The WFprocessor
    also retrieves completed tasks from the TaskManager and updates states of
    PST accordingly.

    :Arguments:
        :sid:             (str)  session id used by the profiler and loggers
        :workflow:        (set)  REFERENCE of the AppManager's workflow
        :resubmit_failed: (bool) True if failed tasks should be resubmitted
        :zmq_info:        (dict) zmq queue addresses
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self,
                 sid,
                 workflow,
                 resubmit_failed,
                 zmq_info):

        # Mandatory arguments
        self._sid             = sid
        self._resubmit_failed = resubmit_failed

        # Assign validated workflow
        self._workflow = workflow

        # Create logger and profiler at their specific locations using the sid
        self._path = os.getcwd() + '/' + self._sid
        self._uid  = ru.generate_id('wfprocessor.%(counter)04d', ru.ID_CUSTOM)

        name = 'radical.entk.%s' % self._uid
        self._logger = ru.Logger  (name, path=self._path)
        self._prof   = ru.Profiler(name, path=self._path)
        self._report = ru.Reporter(name)

        # Defaults
        self._wfp_process       = None
        self._enqueue_thread    = None
        self._dequeue_thread    = None
        self._enqueue_thread_terminate = None
        self._dequeue_thread_terminate = None

        self._logger.info('Created WFProcessor object: %s', self._uid)
        self._prof.prof('create_wfp', uid=self._uid)

        self._setup_zmq(zmq_info)


    # --------------------------------------------------------------------------
    #
    def _setup_zmq(self, zmq_info):

        sid = self._sid
        self._zmq_queue = {
                'put' : ru.zmq.Putter(sid, url=zmq_info['put'], path=sid),
                'get' : ru.zmq.Getter(sid, url=zmq_info['get'], path=sid)}


    # --------------------------------------------------------------------------
    #
    def _advance(self, obj, obj_type, new_state):
        '''
        transition `obj` of type `obj_type` into state `new_state`
        '''

        # NOTE: this is a local operation, no queue communication is involved
        #       (other than the `_advance()` in the TaskManager classes, which

        if   obj_type == 'Task' : msg = obj.parent_stage['uid']
        elif obj_type == 'Stage': msg = obj.parent_pipeline['uid']
        else                    : msg = None

        obj.state = new_state

        self._prof.prof('advance', uid=obj.uid, state=obj.state, msg=msg)
        self._logger.info('Transition %s to state %s', obj.uid, new_state)

        self._report.ok('Update: ')
        self._report.info('%s state: %s\n' % (obj.luid, obj.state))

        if obj_type == 'Task' and obj.state == FAILED:
            self._report.error('task %s failed: %s\n%s\n'
                              % (obj.uid, obj.exception, obj.exception_detail))

        self._logger.info('Transition %s to state %s', obj.uid, new_state)


    # --------------------------------------------------------------------------
    # Getter
    #
    @property
    def workflow(self):
        return self._workflow


    # --------------------------------------------------------------------------
    # Private Methods
    #
    def _create_workload(self):

        # We iterate through all pipelines to collect tasks from
        # stages that are pending scheduling. Once collected, these tasks
        # will be communicated to the tmgr in bulk.

        # Initial empty list to store executable tasks across different
        # pipelines
        workload = list()

        # The executable tasks can belong to different pipelines, and
        # hence different stages. Empty list to store the stages so that
        # we can update the state of stages accordingly
        scheduled_stages = list()

        for pipe in self._workflow:

            with pipe.lock:

                # If Pipeline is in the final state or suspended, we
                # skip processing it.
                if pipe.state in FINAL or  \
                   pipe.completed or \
                   pipe.state == SUSPENDED:
                    continue

                if pipe.state == INITIAL:

                    # Set state of pipeline to SCHEDULING if it is in INITIAL
                    self._advance(pipe, 'Pipeline', SCHEDULING)

                # Get the next stage of this pipeline to process
                exec_stage = pipe.stages[pipe.current_stage - 1]

                if not exec_stage.uid:
                    # TODO: Move parent uid, name assignment to assign_uid()
                    exec_stage.parent_pipeline['uid']  = pipe.uid
                    exec_stage.parent_pipeline['name'] = pipe.name

                # If its a new stage, update its state
                if exec_stage.state == INITIAL:

                    self._advance(exec_stage, 'Stage', SCHEDULING)

                # Get all tasks of a stage in SCHEDULED state
                exec_tasks = list()
                if exec_stage.state == SCHEDULING:
                    exec_tasks = exec_stage.tasks

                for exec_task in exec_tasks:

                    state = exec_task.state
                    if   state == INITIAL or \
                        (state == FAILED and self._resubmit_failed):

                        # Set state of Tasks in current Stage
                        # to SCHEDULING
                        self._advance(exec_task, 'Task', SCHEDULING)

                        # Store the tasks from different pipelines
                        # into our workload list. All tasks will
                        # be submitted in bulk and their states
                        # will be updated accordingly
                        workload.append(exec_task)

                        # We store the stages since the stages the
                        # above tasks belong to also need to be
                        # updated. If its a task that failed, the
                        # stage is already in the correct state
                        if exec_task.state == FAILED:
                            continue

                        if exec_stage not in scheduled_stages:
                            scheduled_stages.append(exec_stage)

        return workload, scheduled_stages


    # --------------------------------------------------------------------------
    #
    def reset_workflow(self):
        '''
        When a component is restarted we reset all the tasks that did not finish
        to the first state. Then they are scheduled again for execution.
        '''

        self._logger.debug('Resetting workflow execution')
        for pipe in self._workflow:

            with pipe.lock:
                if pipe.state in FINAL or  \
                   pipe.completed or \
                   pipe.state == SUSPENDED:
                    continue

                curr_stage = pipe.stages[pipe.current_stage - 1]
                for task in curr_stage.tasks:
                    if task.state not in FINAL:
                        self._advance(task, 'Task', INITIAL)
                self._advance(curr_stage, 'Stage', SCHEDULING)

    # --------------------------------------------------------------------------
    #
    def _execute_workload(self, workload, scheduled_stages):

        msg = {'type': 'workload',
               'body': [task.as_dict() for task in workload]}

        # Send the workload to the pending queue
        self._zmq_queue['put'].put(qname='pending', msgs=[msg])

        self._logger.debug('Workload submitted to Task Manager')

        # Update the state of the tasks in the workload
        for task in workload:

            # Set state of Tasks in current Stage to SCHEDULED
            self._advance(task, 'Task', SCHEDULED)

        # Update the state of the stages from which tasks have
        # been scheduled
        if scheduled_stages:
            for executable_stage in scheduled_stages:
                self._advance(executable_stage, 'Stage', SCHEDULED)


    # --------------------------------------------------------------------------
    #
    def _enqueue(self):
        """
        **Purpose**: This is the function that is run in the enqueue thread.
        This function extracts Tasks from the workflow that exists in
        the WFprocessor object and pushes them to the queues in the pending_q
        list.
        """

        try:

            self._prof.prof('enq_start', uid=self._uid)
            self._logger.info('enqueue-thread started')

            while not self._enqueue_thread_terminate.is_set():

                # Raise an exception while running tests
                ru.raise_on(tag='enqueue_fail')

                workload, scheduled_stages = self._create_workload()

                # If there are tasks to be executed
                if workload:
                    self._execute_workload(workload, scheduled_stages)

            self._logger.info('Enqueue thread terminated')
            self._prof.prof('enq_stop', uid=self._uid)

        except KeyboardInterrupt as ex:

            self._logger.exception('Execution interrupted by user (you \
                                    probably hit Ctrl+C), trying to cancel \
                                    enqueuer thread gracefully...')
            raise KeyboardInterrupt from ex

        except Exception as ex:

            self._logger.exception('Error in enqueue-thread')
            raise EnTKError(ex) from ex



    # --------------------------------------------------------------------------
    #
    def _update_dequeued_task(self, deq_task):

        # Traverse the entire workflow to find out the correct Task
        # TODO: Investigate whether we can change the DS of the
        # workflow so that we don't have this expensive search
        # for each task.
        # First search across all pipelines
        # Note: deq_task is not the same as the task that exists in this process,
        # they are different objects and have different state histories.
        for pipe in self._workflow:

            with pipe.lock:

                # Skip pipelines that have completed or are
                # currently suspended
                if pipe.completed or pipe.state == SUSPENDED:
                    continue

                # Skip pipelines that don't match the UID
                # There will be only one pipeline that matches
                if pipe.uid != deq_task.parent_pipeline['uid']:
                    continue

                self._logger.debug('Found parent pipeline: %s', pipe.uid)

                # Next search across all stages of a matching
                # pipelines
                assert pipe.stages

                stage = None
                for stage in pipe.stages:

                    # Skip stages that don't match the UID
                    # There will be only one stage that matches
                    if stage.uid != deq_task.parent_stage['uid']:
                        continue

                    self._logger.debug('Found parent stage: %s', stage.uid)

                    # Search across all tasks of matching stage
                    for task in stage.tasks:

                        # Skip tasks that don't match the UID
                        # There will be only one task that matches
                        if task.uid != deq_task.uid:
                            continue

                        # If there is no exit code, we assume success
                        # We are only concerned about state of task and not
                        # deq_task
                        if deq_task.exit_code == 0:
                            task_state = DONE
                        elif deq_task.exit_code == 1:
                            task_state = FAILED
                        else:
                            task_state = deq_task.state

                        task.exception        = deq_task.exception
                        task.exception_detail = deq_task.exception_detail

                        if task.state == FAILED and self._resubmit_failed:
                            task_state = INITIAL
                        self._advance(task, 'Task', task_state)

                        # Found the task and processed it -- no more
                        # iterations needed
                        break

                    # Found the stage and processed it -- no more
                    # iterations needed for the current task
                    break

                assert stage

                # Check if current stage has completed
                # If yes, we need to (i) check for post execs to
                # be executed and (ii) check if it is the last
                # stage of the pipeline -- update pipeline
                # state if yes.
                if stage._check_stage_complete():        # pylint: disable=W0212
                    self._advance(stage, 'Stage', DONE)

                    # Check if the current stage has a post-exec
                    # that needs to be executed
                    if stage.post_exec:
                        self._execute_post_exec(pipe, stage)

                    # if pipeline got suspended, advance state accordingly
                    if pipe.state == SUSPENDED:
                        self._advance(pipe, 'Pipeline', SUSPENDED)

                    else:
                        # otherwise perform normal stage progression
                        pipe._increment_stage()          # pylint: disable=W0212

                    # If pipeline has completed, advance state to DONE
                    if pipe.completed:
                        self._advance(pipe, 'Pipeline', DONE)


                # Found the pipeline and processed it -- no more
                # iterations needed for the current task
                break


    # --------------------------------------------------------------------------
    #
    def _execute_post_exec(self, pipe, stage):
        """
        **Purpose**: This method executes the post_exec step of a stage for a
        pipeline.
        """
        try:
            self._logger.info('Executing post-exec for stage %s', stage.uid)
            self._prof.prof('post_exec_start', uid=self._uid)
            resumed_pipe_uids = stage.post_exec()
            self._logger.info('Post-exec executed for stage %s', stage.uid)
            self._prof.prof('post_exec_stop', uid=self._uid)


        except Exception as ex:
            self._logger.exception('post_exec of stage %s failed', stage.uid)
            self._prof.prof('post_exec_fail', uid=self._uid)
            raise EnTKError(ex) from ex

        if resumed_pipe_uids:

            for r_pipe in self._workflow:

                if r_pipe == pipe:
                    continue

                with r_pipe.lock:
                    if r_pipe.uid in resumed_pipe_uids:

                        # Resumed pipelines already have the correct state,
                        # they just need to be synced with the AppMgr.
                        r_pipe._increment_stage()        # pylint: disable=W0212

                        if r_pipe.completed:
                            self._advance(r_pipe, 'Pipeline', DONE)

                        else:
                            self._advance(r_pipe, 'Pipeline', r_pipe.state)



    # --------------------------------------------------------------------------
    #
    def _dequeue(self):
        """
        **Purpose**: This is the function that is run in the dequeue thread.
        This function extracts Tasks from the completed queues and updates the
        workflow.
        """

        try:

            self._prof.prof('deq_start', uid=self._uid)
            self._logger.info('Dequeue thread started')

            while not self._dequeue_thread_terminate.is_set():

                msgs = self._zmq_queue['get'].get_nowait(qname='completed',
                                                         timeout=100)
                if msgs:
                    for msg in msgs:

                        deq_task = Task(from_dict=msg)

                        self._logger.info('Got finished task %s from queue',
                                          deq_task.uid)
                        self._update_dequeued_task(deq_task)


            self._logger.info('Terminated dequeue thread')
            self._prof.prof('deq_stop', uid=self._uid)


        except KeyboardInterrupt as ex:
            self._logger.exception('Execution interrupted by user (you \
                                    probably hit Ctrl+C), trying to exit \
                                    gracefully...')
            raise KeyboardInterrupt from ex


        except Exception as ex:
            self._logger.exception('Error in dequeue-thread')
            raise EnTKError(ex) from ex


        finally:
            self.terminate_processor()


    # --------------------------------------------------------------------------
    #
    # Public Methods
    #
    def start_processor(self):
        """
        **Purpose**: Method to start the wfp process. The wfp function
        is not to be accessed directly. The function is started in a separate
        process using this method.
        """

        try:

            self._logger.info('Starting WFprocessor')
            self._prof.prof('wfp_start', uid=self._uid)

            self._enqueue_thread_terminate = mt.Event()
            self._dequeue_thread_terminate = mt.Event()

            # Start dequeue thread
            self._dequeue_thread = mt.Thread(target=self._dequeue,
                                             name='dequeue-thread')
            self._logger.info('Starting dequeue-thread')
            self._prof.prof('starting dequeue-thread', uid=self._uid)
            self._dequeue_thread.start()

            # Start enqueue thread
            self._enqueue_thread = mt.Thread(target=self._enqueue,
                                             name='enqueue-thread')
            self._logger.info('Starting enqueue-thread')
            self._prof.prof('starting enqueue-thread', uid=self._uid)
            self._enqueue_thread.start()

            self._logger.info('WFprocessor started')
            self._prof.prof('wfp_started', uid=self._uid)

        except Exception as ex:

            self._logger.exception('WFprocessor failed')
            self.terminate_processor()
            raise EnTKError(ex) from ex


    # --------------------------------------------------------------------------
    #
    def terminate_processor(self):
        """
        **Purpose**: Method to terminate the wfp process. This method is
        blocking as it waits for the wfp process to terminate (aka join).
        """

        try:

            tid = mt.current_thread().ident

            if self._enqueue_thread:

                if not self._enqueue_thread_terminate.is_set():
                    self._logger.info('Terminating enqueue-thread')
                    self._enqueue_thread_terminate.set()
                    if tid != self._enqueue_thread.ident:
                        self._enqueue_thread.join()
                    self._enqueue_thread = None

            if self._dequeue_thread:

                if not self._dequeue_thread_terminate.is_set():
                    self._logger.info('Terminating dequeue-thread')
                    self._dequeue_thread_terminate.set()
                    if tid != self._dequeue_thread.ident:
                        self._dequeue_thread.join()
                    self._dequeue_thread = None

            self._logger.info('WFprocessor terminated')
            self._prof.prof('wfp_stop', uid=self._uid)
            self._prof.close()

        except Exception as ex:

            self._logger.exception('Could not terminate wfprocessor process')
            raise EnTKError(ex) from ex


    # --------------------------------------------------------------------------
    #
    def workflow_incomplete(self):
        """
        **Purpose**: Method to check if the workflow execution is incomplete.
        """

        try:
            for pipe in self._workflow:
                with pipe.lock:
                    if pipe.completed:
                        pass
                    else:
                        return True
            return False

        except Exception as ex:
            self._logger.exception(
                'Could not check if workflow is incomplete, error:%s' % ex)
            raise EnTKError(ex) from ex


    # --------------------------------------------------------------------------
    #
    def check_processor(self):

        if self._enqueue_thread and \
           self._dequeue_thread and \
           self._enqueue_thread.is_alive() and \
           self._dequeue_thread.is_alive():
            return True

        return False


# ------------------------------------------------------------------------------




#!/usr/bin/env python

from radical.entk import Pipeline, Stage, Task, AppManager
import os

# ------------------------------------------------------------------------------
# Set default verbosity
if os.environ.get('RADICAL_ENTK_VERBOSE') is None:
    os.environ['RADICAL_ENTK_REPORT'] = 'True'


if __name__ == '__main__':

    # Create a Pipeline object
    p = Pipeline()

    # Create a Stage object
    s1 = Stage()
    s1.name = 'Stage.1'

    for cnt in range(10):

        # Create a Task object
        t = Task()
        t.name = 'my.task'        # Assign a name to the task (optional)
        t.executable = '/bin/echo'   # Assign executable to the task
        t.arguments = ['I am task %s in %s' % (cnt, s1.name)]  # Assign arguments for the task executable

        # Add the Task to the Stage
        s1.add_tasks(t)

    # Add Stage to the Pipeline
    p.add_stages(s1)


    # Create another Stage object
    s2 = Stage()
    s2.name = 'Stage.2'

    for cnt in range(5):

        # Create a Task object
        t = Task()
        t.name = 'my.task'        # Assign a name to the task (optional, do not use ',' or '_')
        t.executable = '/bin/echo'   # Assign executable to the task
        t.arguments = ['I am task %s in %s' % (cnt, s2.name)]  # Assign arguments for the task executable

        # Add the Task to the Stage
        s2.add_tasks(t)

    # Add Stage to the Pipeline
    p.add_stages(s2)


    # Create Application Manager
    appman = AppManager()

    # Create a dictionary describe four mandatory keys:
    # resource, walltime, and cpus
    # resource is 'local.localhost' to execute locally
    res_dict = {

        'resource': 'local.localhost',
        'walltime': 10,
        'cpus': 1
    }

    # Assign resource request description to the Application Manager
    appman.resource_desc = res_dict

    # Assign the workflow as a set or list of Pipelines to the Application Manager
    # Note: The list order is not guaranteed to be preserved
    appman.workflow = set([p])

    # Run the Application Manager
    appman.run()




__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__    = "Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>"
__license__   = "MIT"

import os

import radical.utils as ru

from ...exceptions import EnTKMissingError, EnTKTypeError, EnTKError


# ------------------------------------------------------------------------------
#
class Base_ResourceManager(object):
    """
    A resource manager takes the responsibility of placing resource requests on
    different, possibly multiple, DCIs.

    :arguments:
        :resource_desc: dictionary with details of the resource request
                        and access credentials of the user
        :example: resource_desc = {
                                    |  'resource'      : 'xsede.stampede',
                                    |  'walltime'      : 120,
                                    |  'cpus'          : 64,
                                    |  'gpus'          : 0,          # optional
                                    |  'project'       : 'TG-abcxyz',
                                    |  'memory'        : '16000'     # optional
                                    |  'queue'         : 'abc',      # optional
                                    |  'access_schema' : 'ssh'       # optional
                                    |  'job_name'      : 'test_job'  # optional}
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, resource_desc, sid, rts, rts_config):

        if not isinstance(resource_desc, dict):
            raise EnTKTypeError(expected_type=dict, actual_type=type(resource_desc))

        self._resource_desc = resource_desc
        self._sid           = sid
        self._rts           = rts
        self._rts_config    = rts_config

        # Resource reservation related parameters
        self._resource      = None
        self._walltime      = None
        self._cpus          = 1
        self._gpus          = 0
        self._memory        = 0
        self._project       = None
        self._access_schema = None
        self._queue         = None
        self._job_name      = None
        self._services      = []
        self._validated     = False

        # Utility parameters
        self._uid = ru.generate_id('resource_manager.%(counter)04d',
                                   ru.ID_CUSTOM)
        self._path = os.getcwd() + '/' + self._sid

        name = 'radical.entk.%s' % self._uid
        self._logger = ru.Logger  (name, path=self._path)
        self._prof   = ru.Profiler(name, path=self._path)

        self._shared_data = list()
        self._outputs     = None


    # --------------------------------------------------------------------------
    #
    @property
    def resource(self):
        """
        :getter: Return user specified resource name
        """
        return self._resource


    @property
    def walltime(self):
        """
        :getter: Return user specified walltime
        """
        return self._walltime


    @property
    def cpus(self):
        """
        :getter: Return user specified number of cpus
        """
        return self._cpus


    @property
    def memory(self):
        """
        :getter: Return user specified amount of memory
        """
        return self._memory


    @property
    def gpus(self):
        """
        :getter: Return user specified number of gpus
        """
        return self._gpus


    @property
    def project(self):
        """
        :getter: Return user specified project ID
        """
        return self._project


    @property
    def access_schema(self):
        """
        :getter:    Return user specified access schema -- 'ssh' or 'gsissh' or
                    None
        """
        return self._access_schema


    @property
    def queue(self):
        """
        :getter: Return user specified resource queue to be used
        """
        return self._queue


    @property
    def job_name(self):
        """
        :getter:    Return user specified job_name
        """
        return self._job_name


    @property
    def services(self):
        """
        :getter: Returns the list of tasks used to start "global" services
        :setter: Assigns a list of service tasks, which are launched before
        any stage starts and run during the whole workflow execution
        """
        return self._services


    @property
    def shared_data(self):
        """
        :getter:    list of files to be staged to remote and that are common to
                    multiple tasks
        :setter:    Assign a list of names of files that need to be accessible to
                    tasks
        """
        return self._shared_data


    @property
    def outputs(self):
        """
        :getter:    list of files to be staged from remote after execution
        :setter:    Assign a list of names of files that need to be staged from the
                    remote machine
        """
        return self._outputs


    # --------------------------------------------------------------------------
    # Setter functions
    #
    @shared_data.setter
    def shared_data(self, data_list):

        self._shared_data = data_list


    @outputs.setter
    def outputs(self, data):

        self._outputs = data


    @services.setter
    def services(self, tasks):

        self._services = tasks


    # --------------------------------------------------------------------------
    #
    def get_resource_allocation_state(self):
        """
        **Purpose**: Get the state of the resource allocation
        """
        raise NotImplementedError('get_resource_allocation_state() method not '
                                  'implemented in ResourceManager for %s'
                                  % self._rts)


    # --------------------------------------------------------------------------
    #
    def get_completed_states(self):
        """
        **Purpose**: Test if a resource allocation was submitted
        """

        raise NotImplementedError('completed_states() method not implemented '
                                  'in ResourceManager for %s' % self._rts)


    # --------------------------------------------------------------------------
    #
    def get_rts_info(self):
        """
        **Purpose**: Return the RTS information as a dict.
        """

        raise NotImplementedError('get_rts_info() method not implemented '
                                  'in ResourceManager for %s' % self._rts)


    # --------------------------------------------------------------------------
    #
    def _validate_resource_desc(self):
        """
        **Purpose**: Validate the provided resource description
        """

        self._prof.prof('rdesc_validate', uid=self._uid)
        self._logger.debug('Validating resource description')

        expected_keys = ['resource',
                         'walltime',
                         'cpus']

        for key in expected_keys:
            if key not in self._resource_desc:
                raise EnTKMissingError(obj='resource description',
                                       missing_attribute=key)

        if not isinstance(self._resource_desc['resource'], str):
            raise EnTKTypeError(expected_type=str,
                            actual_type=type(self._resource_desc['resource']))

        if not isinstance(self._resource_desc['walltime'], int):
            raise EnTKTypeError(expected_type=int,
                            actual_type=type(self._resource_desc['walltime']))

        if not isinstance(self._resource_desc['cpus'], int):
            raise EnTKTypeError(expected_type=int,
                            actual_type=type(self._resource_desc['cpus']))

        if 'memory' in self._resource_desc:
            if not isinstance(self._resource_desc['memory'], int):
                raise EnTKTypeError(expected_type=int,
                                actual_type=type(self._resource_desc['memory']))

        if 'gpus' in self._resource_desc:
            if not isinstance(self._resource_desc['gpus'], int):
                raise EnTKTypeError(expected_type=int,
                                actual_type=type(self._resource_desc['gpus']))

        if 'project' in self._resource_desc:
            if not isinstance(self._resource_desc['project'], str):
                raise EnTKTypeError(expected_type=str,
                               actual_type=type(self._resource_desc['project']))

        if 'access_schema' in self._resource_desc:
            if not isinstance(self._resource_desc['access_schema'], str):
                raise EnTKTypeError(expected_type=str,
                         actual_type=type(self._resource_desc['access_schema']))

        if 'queue' in self._resource_desc:
            if not isinstance(self._resource_desc['queue'], str):
                raise EnTKTypeError(expected_type=str,
                                actual_type=type(self._resource_desc['queue']))

        if not isinstance(self._rts_config, dict):
            raise EnTKTypeError(expected_type=dict,
                            actual_type=type(self._rts_config))

        self._logger.info('Resource description validated')
        self._prof.prof('rdesc_valid', uid=self._uid)

        self._validated = True

        return self._validated


    # --------------------------------------------------------------------------
    #
    def _populate(self):
        """
        **Purpose**:    Populate the ResourceManager class with the validated
                        resource description
        """

        if not self._validated:
            raise EnTKError('Resource description not validated')

        self._prof.prof('populating rmgr', uid=self._uid)
        self._logger.debug('Populating resource manager object')

        self._resource      = self._resource_desc['resource']
        self._walltime      = self._resource_desc['walltime']
        self._cpus          = self._resource_desc['cpus']
        self._memory        = self._resource_desc.get('memory',        0)
        self._gpus          = self._resource_desc.get('gpus',          0)
        self._project       = self._resource_desc.get('project',       None)
        self._access_schema = self._resource_desc.get('access_schema', None)
        self._queue         = self._resource_desc.get('queue',         None)
        self._job_name      = self._resource_desc.get('job_name',      None)

        self._logger.debug('Resource manager population successful')
        self._prof.prof('rmgr populated', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def submit_resource_request(self):
        """
        **Purpose**:  Submit resource request per provided description
        """

        raise NotImplementedError('submit_resource_request() method not '
                            'implemented in ResourceManager for %s' % self._rts)


    # --------------------------------------------------------------------------
    #
    def _terminate_resource_request(self):
        """
        **Purpose**:  Cancel resource request by terminating any reservation
                      on any acquired resources or resources pending acquisition
        """

        raise NotImplementedError('_terminate_resource_request() method not '
                            'implemented in ResourceManager for %s' % self._rts)


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2023, The RADICAL-Cybertools Team'
__license__   = 'MIT'

from typing import Optional, Dict, List

import radical.utils as ru

from .. import Pipeline


# ------------------------------------------------------------------------------
#
def get_provenance_graph(pipelines: List[Pipeline],
                         output_file: Optional[str] = None) -> Dict:
    """
    Using UIDs of all entities to build a workflow provenance graph.
    """

    graph = {}

    pipelines = ru.as_list(pipelines)

    for pipeline in pipelines:
        graph[pipeline.uid] = {}

        for stage in pipelines.stages:
            graph[pipeline.uid][stage.uid] = {}

            for task in stage.tasks:
                g_task = graph[pipeline.uid][stage.uid].setdefault(task.uid, {})
                if task.annotations:
                    g_task.update(task.annotations.as_dict())

    if output_file:
        if not output_file.endswith('.json'):
            output_file += '.json'
        ru.write_json(graph, output_file)

    return graph


# ------------------------------------------------------------------------------
#
def extract_provenance_graph(session_json: str,
                             output_file: Optional[str] = None) -> Dict:
    """
    Using session JSON file to build a workflow provenance graph.
    """

    session_entities = ru.read_json(session_json)

    if not session_entities.get('task'):
        raise ValueError('No task entities in provided session')

    graph = {}

    for task in session_entities['task']:
        task_uid, _, stage_uid, _, pipeline_uid, _ = task['name'].split(',')
        graph.\
            setdefault(pipeline_uid, {}).\
            setdefault(stage_uid, {}).\
            setdefault(task_uid,
                       task['description']['metadata'].get('data') or {})

    for pipeline_uid in graph:
        for stage_uid in graph[pipeline_uid]:
            graph[pipeline_uid][stage_uid].sort()

    if output_file:
        if not output_file.endswith('.json'):
            output_file += '.json'
        ru.write_json(graph, output_file)

    return graph


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2017-2018, http://radical.rutgers.edu'
__author__    = 'Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>'
__license__   = 'MIT'

import os
import time
import warnings

import threading       as mt
import multiprocessing as mp

import radical.utils   as ru

from ..exceptions import EnTKError, EnTKTypeError, EnTKMissingError
from ..states     import DONE, FAILED

from ..pipeline   import Pipeline
from ..task       import Task
from ..utils      import write_session_description
from ..utils      import write_workflows

from .wfprocessor import WFprocessor


# pylint: disable=protected-access

# ------------------------------------------------------------------------------
#
class AppManager(object):

    '''
    An application manager takes the responsibility of setting up the
    communication infrastructure, instantiates the ResourceManager, TaskManager,
    WFProcessor objects and all their threads and processes. This is the Master
    object running in the main process and is designed to recover from errors
    from all other objects, threads and processes.

    :Arguments:
        :config_path:     Url to config path to be read for AppManager
        :reattempts:      number of attempts to re-invoke any failed EnTK
                          components
        :resubmit_failed: resubmit failed tasks (True/False)
        :autoterminate:   terminate resource reservation upon execution of all
                          tasks of first workflow (True/False)
        :write_workflow:  write workflow and mapping to rts entities to a file
                          (post-termination)
        :rts:             Specify RTS to use. Current options: 'mock',
                          'radical.pilot' (default if unspecified)
        :rts_config:      Configuration for the RTS, accepts
                          {'sandbox_cleanup': True/False,'db_cleanup':
                          True/False} when RTS is RP
        :name:            Name of the Application. It should be unique between
                          executions. (default is randomly assigned)
        :base_path:       base path of logger and profiler with a session id,
                          None (default) will use a current working directory
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self,
                 config_path=None,
                 reattempts=None,
                 resubmit_failed=None,
                 autoterminate=None,
                 write_workflow=None,
                 rts=None,
                 rts_config=None,
                 name=None,
                 base_path=None,
                 **kwargs):

        for arg in ['hostname', 'port', 'username', 'password']:
            if arg in kwargs:
                warnings.warn('%s argument is not required anymore' % arg,
                              DeprecationWarning, stacklevel=2)

        # Create a session for each EnTK script execution
        if name:
            self._name = name
            self._sid  = name
        else:
            self._name = str()
            self._sid  = ru.generate_id('re.session', ru.ID_PRIVATE)

        self._read_config(config_path, reattempts, resubmit_failed,
                          autoterminate, write_workflow, rts, rts_config,
                          base_path)

        # Create an uid + logger + profiles for AppManager, under the sid
        # namespace

        self._uid = ru.generate_id('appmanager.%(counter)04d', ru.ID_CUSTOM)

        path = str(self._base_path) + '/' + self._sid
        name = 'radical.entk.%s' % self._uid

        self._logger = ru.Logger(name=name, path=path)
        self._prof   = ru.Profiler(name=name, path=path)
        self._report = ru.Reporter(name=name)

        self._report.info('EnTK session: %s\n' % self._sid)
        self._report.info('Creating AppManager\n')
        self._prof.prof('amgr_creat', uid=self._uid)

        self._rmgr            = None

        # Global parameters to have default values
        self._resource_desc   = None
        self._services        = list()
        self._task_manager    = None
        self._workflow        = None
        self._workflows       = list()
        self._cur_attempt     = 1
        self._shared_data     = list()
        self._outputs         = None
        self._wfp             = None
        self._sync_thread     = None
        self._terminate_sync  = mt.Event()
        self._resubmit_failed = False
        self._term            = mp.Event()

        # Setup zmq queues
        self._zmq_info        = dict()
        self._zmq_queue       = None
        self._zmq_bridge      = None
        self._setup_zmq()

        self._prof.prof('amgr_created', uid=self._uid)
        self._report.info('AppManager initialized')
        self._report.ok('>>ok\n')


    # --------------------------------------------------------------------------
    #
    def _read_config(self, config_path, reattempts, resubmit_failed,
                           autoterminate, write_workflow, rts, rts_config,
                           base_path):

        if not config_path:
            config_path = os.path.dirname(os.path.abspath(__file__))

        config = ru.read_json(os.path.join(config_path, 'config.json'))

        def _if(val1, val2):
            if val1 is not None: return val1
            else               : return val2

        self._base_path        = _if(base_path,       os.getcwd())
        self._reattempts       = _if(reattempts,      config['reattempts'])
        self._resubmit_failed  = _if(resubmit_failed, config['resubmit_failed'])
        self._autoterminate    = _if(autoterminate,   config['autoterminate'])
        self._write_workflow   = _if(write_workflow,  config['write_workflow'])
        self._rts_config       = _if(rts_config,      config['rts_config'])
        self._rts              = _if(rts,             config['rts'])

        if self._rts not in ['radical.pilot', 'mock']:
            raise ValueError('invalid RTS %s' % self._rts)


    # --------------------------------------------------------------------------
    #
    # Getter functions
    #
    @property
    def name(self):
        '''
        Name for the application manager. Allows the user to setup the name of
        the application manager, as well as its session ID. This name should be
        unique between different EnTK executions, otherwise it will produce an
        error.

        :getter: Returns the name of the application manager
        :setter: Assigns the name of the application manager
        :type: String
        '''

        return self._name


    # --------------------------------------------------------------------------
    #
    @property
    def sid(self):
        '''
        Get the session ID of the current EnTK execution

        :getter: Returns the session ID of the EnTK execution
        :type: String
        '''

        return self._sid


    # --------------------------------------------------------------------------
    #
    @property
    def resource_desc(self):
        '''
        The resource description is a dictionary that holds information about
        the resource(s) that will be used to execute the workflow.

        The following keys are mandatory in all resource descriptions:
            | 'resource'      : Label of the platform with resources.
            | 'walltime'      : Amount of time the workflow is expected to run.
            | 'cpus'          : Number of CPU cores/threads.

        Optional keys include:
            | 'project'       : The project that will be charged.
            | 'gpus'          : Number of GPUs to be used by the workflow.
            | 'access_schema' : The key of an access mechanism to use.
            | 'queue'         : The name of the batch queue.
            | 'job_name'      : The specific name of a batch job.

        :getter: Returns the resource description
        :setter: Assigns a resource description
        '''

        return self._resource_desc


    # --------------------------------------------------------------------------
    #
    @property
    def services(self):
        '''
        :getter: Returns the list of tasks used to start "global" services
        :setter: Assigns a list of service tasks, which are launched before
        any stage starts and run during the whole workflow execution
        '''

        return self._rmgr.services if self._rmgr else self._services


    # --------------------------------------------------------------------------
    #
    @property
    def workflow(self):
        '''
        :getter: Return the last workflow assigned for execution
        :setter: Assign a new workflow to be executed
        '''

        return self._workflow


    # --------------------------------------------------------------------------
    #
    @property
    def workflows(self):
        """
        :getter: Return a list of workflows assigned for execution
        """

        return self._workflows


    # --------------------------------------------------------------------------
    #
    @property
    def shared_data(self):
        '''
        :getter: Return list of filenames that are shared between multiple tasks
                 of the application
        :setter: Assign a list of names of files that need to be staged to the
                remote machine
        '''

        return self._shared_data


    @property
    def outputs(self):
        '''
        :getter: Return list of filenames that are to be staged out after
                 execution
        :setter: Assign a list of names of files that need to be staged from the
                 remote machine
        '''

        return self._outputs


    # --------------------------------------------------------------------------
    # Setter functions
    #
    @name.setter
    def name(self, value):

        if not isinstance(value, str):
            raise EnTKTypeError(expected_type=str, actual_type=type(value))

        self._name = value


    # --------------------------------------------------------------------------
    #
    @resource_desc.setter
    def resource_desc(self, value):

        if self._rts == 'radical.pilot':
            from ..execman.rp import ResourceManager

        elif self._rts == 'mock':
            from ..execman.mock import ResourceManager

        else:
            raise ValueError('unknown RTS %s' % self._rts)

        self._rmgr = ResourceManager(resource_desc=value, sid=self._sid,
                                     rts_config=self._rts_config)

        self._report.info('Validating and assigning resource manager')

        if not self._rmgr._validate_resource_desc():
            self._logger.error('Could not validate resource description')
            raise EnTKError('Could not validate resource description')

        self._resource_desc = value

        self._rmgr._populate()
        self._rmgr.shared_data = self._shared_data
        self._rmgr.outputs     = self._outputs

        if self._services:
            self._rmgr.services = self._services

        self._report.ok('>>ok\n')


    # --------------------------------------------------------------------------
    #
    @services.setter
    def services(self, tasks):

        if not isinstance(tasks, list):
            tasks = [tasks]

        for task in tasks:
            if not isinstance(task, Task):
                raise EnTKTypeError(expected_type=Task,
                                    actual_type=type(task))

        self._services = tasks
        if self._rmgr:
            self._rmgr.services = tasks


    # --------------------------------------------------------------------------
    #
    @workflow.setter
    def workflow(self, workflow):

        self._prof.prof('assigning workflow', uid=self._uid)

        for p in workflow:

            if not isinstance(p, Pipeline):
                self._logger.info('workflow type incorrect')
                raise EnTKTypeError(expected_type=['Pipeline',
                                                   'set of Pipelines'],
                                    actual_type=type(p))
            p._validate()

        # keep history
        self._workflows.append(workflow)

        # set current workflow
        self._workflow = workflow
        self._logger.info('Workflow assigned to Application Manager')


    # --------------------------------------------------------------------------
    #
    @shared_data.setter
    def shared_data(self, data):

        if not isinstance(data, list):
            data = [data]

        for value in data:
            if not isinstance(value, str):
                raise EnTKTypeError(expected_type=str,
                                    actual_type=type(value))

        self._shared_data = data

        if self._rmgr:
            self._rmgr.shared_data = data



    # --------------------------------------------------------------------------
    #
    @outputs.setter
    def outputs(self, data):

        if not isinstance(data, list):
            data = [data]

        for value in data:
            if not isinstance(value, str):
                raise EnTKTypeError(expected_type=str,
                                    actual_type=type(value))

        if self._rmgr:
            self._rmgr.outputs = data


    # --------------------------------------------------------------------------
    #
    # Public methods
    #
    def run(self):
        '''
        **Purpose**: Run the application manager. Once the workflow and resource
        manager have been assigned. Invoking this method will start the setting
        up the communication infrastructure, submitting a resource request and
        then submission of all the tasks.
        '''

        ret = None

        try:
            self._prof.prof('amgr_start', uid=self._uid)

            # Set None for variables local to each run
            self._resubmit_failed = False
            self._cur_attempt = 1

            # Ensure that a workflow and a resource description have
            # been defined
            if not self._workflow:
                self._logger.error('No workflow assigned currently, please \
                                    check your script')
                raise EnTKMissingError(obj=self._uid,
                                       missing_attribute='workflow')

            if not self._rmgr:
                self._logger.error('No resource manager assigned currently, \
                                    please create and add a valid resource \
                                    manager')
                raise EnTKMissingError(obj=self._uid,
                                       missing_attribute='resource_manager')
            self._prof.prof('amgr run started', uid=self._uid)

            # ensure zmq setup
            self._setup_zmq()

            # Submit resource request if no resource allocation done till now or
            # resubmit a new one if the old one has completed
            res_alloc_state = self._rmgr.get_resource_allocation_state()
            if not res_alloc_state or \
                   res_alloc_state in self._rmgr.get_completed_states():

                self._logger.info('Starting resource request submission')
                self._prof.prof('rreq_init', uid=self._uid)

                self._rmgr.submit_resource_request()

                res_alloc_state = self._rmgr.get_resource_allocation_state()
                if res_alloc_state in self._rmgr.get_completed_states():
                    raise EnTKError('Cannot proceed. Resource ended in state %s'
                                    % res_alloc_state)


            # Start all components and subcomponents
            self._start_all_comps()

            # Run workflow -- this call is blocking till all tasks of the
            # workflow are executed or an error/exception is encountered
            self._run_workflow()
            self._logger.info('Workflow execution finished.')
            if self._autoterminate:
                self._logger.debug('Autoterminate set to %s',
                                   self._autoterminate)
                self.terminate()

        except KeyboardInterrupt as ex:

            self._logger.exception('Execution interrupted by user (you '
                                   'probably hit Ctrl+C), trying to cancel '
                                   'enqueuer thread gracefully...')
            self.terminate()
            raise EnTKError(ex) from ex

        except Exception as ex:

            self._logger.exception('Error in AppManager')
            self.terminate()
            raise EnTKError(ex) from ex

        # return list of fetched output data, or None.
        outputs = self.outputs
        if outputs:
            ret = outputs
        return ret


    # --------------------------------------------------------------------------
    #
    def terminate(self):

        self._prof.prof('term_start', uid=self._uid)

        self._term.set()

        # Terminate threads in following order: wfp, helper, synchronizer
        if self._wfp:
            self._logger.info('Terminating WFprocessor')
            self._wfp.terminate_processor()

        if self._task_manager:
            self._logger.info('Terminating task manager process')
            self._task_manager.terminate_manager()

        if self._sync_thread:
            self._logger.info('Terminating synchronizer thread')
            self._terminate_sync.set()
            self._sync_thread.join()
            self._logger.info('Synchronizer thread terminated')

        if self._write_workflow:
            write_workflows(self.workflows, self._sid)

        if self._rmgr:
            self._rmgr._terminate_resource_request()

        if os.environ.get('RADICAL_ENTK_PROFILE'):
            write_session_description(self)

        self._report.info('All components terminated\n')
        self._prof.prof('termination done', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def resource_terminate(self):

        self._logger.warning('DeprecationWarning: `resource_terminate()` is '
                             'deprecated, please use `terminate()`')
        self.terminate()


    # --------------------------------------------------------------------------
    #
    # Private methods
    #
    def _setup_zmq(self):
        '''
        **Purpose**: Setup ZMQ system on the client side. We instantiate
        queue(s) 'pendingq-*' for communication between the enqueuer thread and
        the task manager process. We instantiate queue(s) 'completedq-*' for
        communication between the task manager and dequeuer thread. We
        instantiate queue 'sync-to-master' for communication from
        enqueuer/dequeuer/task_manager to the synchronizer thread. We
        instantiate queue 'sync-ack' for communication from synchronizer thread
        to enqueuer/dequeuer/task_manager.
        '''

        try:
            sid = self._sid
            self._report.info('Setting up ZMQ queues')

            if self._zmq_info:
                self._report.ok('>>n/a\n')
                return

            self._report.ok('>>ok\n')

            self._prof.prof('zmq_setup_start', uid=self._uid)
            self._logger.debug('Setting up zmq queues')

            cfg = ru.Config(cfg={'channel'   : sid,
                                 'uid'       : sid,
                                 'path'      : sid,
                                 'type'      : 'queue',
                                 'stall_hwm' : 0,
                                 'bulk_size' : 1})

            self._zmq_bridge = ru.zmq.Queue(sid, cfg)
            self._zmq_bridge.start()
            time.sleep(1)

            zmq_info = {
                    'put': str(self._zmq_bridge.addr_put),
                    'get': str(self._zmq_bridge.addr_get)}

            self._zmq_queue = {
                    'put' : ru.zmq.Putter(sid, url=zmq_info['put'], path=sid),
                    'get' : ru.zmq.Getter(sid, url=zmq_info['get'], path=sid)}

            self._zmq_info = zmq_info


        except Exception as e:

            self._logger.exception('Error setting ZMQ queues')
            raise EnTKError(e) from e


    # --------------------------------------------------------------------------
    #
    def _start_all_comps(self):

        if self._wfp:
            # This condition is called when there are multiple workflows
            # submitted for execution. Amgr.run() was probably called twice.
            # If a WFP exists, we use the same one but with the new workflow.
            # Since WFP (and its threads) and the Amgr share memory, we have
            # to terminate WFP's threads, assign the new workflow and then
            # start the threads again.
            self._wfp.terminate_processor()
            self._wfp._workflow = self._workflow
            self._wfp.start_processor()
            return

        # Create WFProcessor and initialize workflow its contents with
        # uids
        self._prof.prof('wfp_create_start', uid=self._uid)
        self._wfp = WFprocessor(sid=self._sid,
                                workflow=self._workflow,
                                resubmit_failed=self._resubmit_failed,
                                zmq_info=self._zmq_info)
        self._prof.prof('wfp_create_stop', uid=self._uid)

        # Start synchronizer thread AM OK
        if not self._sync_thread:
            self._logger.info('Starting synchronizer thread')
            self._sync_thread = mt.Thread(target=self._synchronizer,
                                          name='synchronizer-thread')
            self._prof.prof('sync_thread_create', uid=self._uid)
            self._sync_thread.start()

        # Start WFprocessor
        self._logger.info('Starting WFProcessor')
        self._wfp.start_processor()
        self._report.ok('All components created\n')

        # Create tmgr object only if it does not already exist
        if self._rts == 'radical.pilot':
            from ..execman.rp import TaskManager

        elif self._rts == 'mock':
            from ..execman.mock import TaskManager

        else:
            raise ValueError('unknown RTS %s' % self._rts)

        if not self._task_manager:

            self._logger.info('Starting task manager')
            self._prof.prof('tmgr_create_start', uid=self._uid)

            self._task_manager = TaskManager(
                    sid=self._sid,
                    rmgr=self._rmgr,
                    zmq_info=self._zmq_info)

            self._task_manager.start_manager()
            self._submit_rts_tmgr(self._rmgr.get_rts_info())
            self._prof.prof('tmgr_create_stop', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def _submit_rts_tmgr(self, rts_info):
        '''
        **Purpose**: Update the runtime system information in the task manager
        '''
        rts_msg = {'type': 'rts',
                   'body': rts_info}

        self._zmq_queue['put'].put(qname='pending', msgs=rts_msg)


    # --------------------------------------------------------------------------
    #
    def _run_workflow(self):

        active_pipe_count  = len(self._workflow or [])
        finished_pipe_uids = list()

        # We wait till all pipelines of the workflow are marked
        # complete
        # incomplete = self._wfp.workflow_incomplete()
        rts_final_states = self._rmgr.get_completed_states()

        while active_pipe_count and self._cur_attempt <= self._reattempts:

            if self._term.is_set():
                break

            for pipe in self._workflow:

                with pipe.lock:

                    if pipe.completed and \
                       pipe.uid not in finished_pipe_uids:

                        finished_pipe_uids.append(pipe.uid)
                        active_pipe_count -= 1

                        self._logger.info('Pipe %s completed', pipe.uid)
                        self._logger.info('Active pipes %s', active_pipe_count)

            reset_workflow = False

            if not self._sync_thread.is_alive():
                self._logger.info('Synchronizer thread is not alive.')
                self._sync_thread = mt.Thread(target=self._synchronizer,
                                              name='synchronizer-thread')
                self._sync_thread.start()
                self._prof.prof('sync_thread_restart', uid=self._uid)
                self._logger.info('Restarted synchronizer thread.')

                reset_workflow = True

            if not self._wfp.check_processor():
                self._logger.info('WFP is not alive.')

                # If WFP dies, both child threads are also cleaned out.
                # We simply recreate the wfp object with a copy of the
                # workflow in the appmanager and start the processor.

                self._prof.prof('wfp_recreate', uid=self._uid)
                self._wfp.terminate_processor()
                self._wfp.start_processor()
                self._logger.info('Restarted WFProcessor.')

                reset_workflow = True

            state = self._rmgr.get_resource_allocation_state()
            if state in rts_final_states:

                self._cur_attempt += 1
                if self._cur_attempt > self._reattempts:
                    break

                self._logger.debug('Workflow not done. Resubmitting RTS.')

                self._rmgr.submit_resource_request()
                rts_info = self._rmgr.get_rts_info()
                self._submit_rts_tmgr(rts_info=rts_info)
                self._logger.debug('RTS resubmitted')

                reset_workflow = True

            if reset_workflow:
                self._wfp.reset_workflow()

        if self._cur_attempt > self._reattempts:
            raise EnTKError('Too many failures in synchronizer, wfp or '
                            'task manager')


    # --------------------------------------------------------------------------
    #
    def _get_message_to_sync(self, qname):
        '''
        Reads a message from the queue, and exchange the message to where it
        was published by `update_task`
        '''

        # --------------------------------------------------------------
        # Messages between tmgr Main thread and synchronizer -- only
        # Task objects
        msgs = self._zmq_queue['get'].get_nowait(qname=qname)

        if not msgs:
            return 0

        # The message received has the following
        # structure:
        # msg = {
        #         'type': 'Pipeline'/'Stage'/'Task',
        #         'object': json/dict
        #         }
        for msg in msgs:

            uid   = msg['object']['uid']
            state = msg['object']['state']

            obj = msg['object']

            self._prof.prof('sync_recv_obj_state_%s' % state, uid=uid)
            self._logger.debug('recv %s in state %s (sync)', uid, state)

            if msg['type'] == 'Task':
                self._update_task(obj)

        return 1


    # --------------------------------------------------------------------------
    #
    def _update_task(self, tdict):
        # pylint: disable=W0612,W0613

        completed_task = Task(from_dict=tdict)

        self._logger.info('Received %s with state %s', completed_task.uid,
                completed_task.state)

        # Traverse the entire workflow to find the correct task
        for pipe in self._workflow:

            with pipe.lock:

                if pipe.completed:
                    continue

                if pipe.uid != completed_task.parent_pipeline['uid']:
                    continue

                for stage in pipe.stages:

                    if stage.uid != completed_task.parent_stage['uid']:
                        continue

                    for task in stage.tasks:

                        if completed_task.uid != task.uid:
                            continue

                        if completed_task.state == task.state:
                            continue

                        self._logger.debug('Found task %s in state (%s) \
                                           changing to %s ==', task.uid,
                                           task.state, completed_task.state)

                        if completed_task.path:
                            task.path = str(completed_task.path)
                            self._logger.debug('Task %s path set to %s',
                                               task.uid, task.path)

                        if completed_task.rts_uid:
                            task.rts_uid = str(completed_task.rts_uid)
                            self._logger.debug('Task %s rts_uid set to %s',
                                               task.uid, task.rts_uid)

                        if task.state in [DONE, FAILED]:
                            self._logger.debug('No change on task state %s \
                                             in state %s', task.uid, task.state)
                            break

                        task.state            = completed_task.state
                        task.exception        = completed_task.exception
                        task.exception_detail = completed_task.exception_detail

                        self._logger.debug('Found task %s in state %s',
                                           task.uid, task.state)

                        state = tdict['state']
                        self._prof.prof('pub_ack_state_%s' % state,
                                        uid=tdict['uid'])

                        self._report.ok('Update: ')
                        self._report.info('%s state: %s\n'
                                         % (task.luid, task.state))

                        break


    # --------------------------------------------------------------------------
    #
    def _synchronizer(self):

        try:
            self._synchronizer_work()

        except KeyboardInterrupt:
            self._logger.exception('Execution interrupted by user (you \
                                    probably hit Ctrl+C), trying to terminate \
                                    synchronizer thread gracefully...')
            raise

        except Exception as ex:
            self._logger.exception('Unknown error in synchronizer: %s. \
                                    Terminating thread')
            raise EnTKError(ex) from ex


    # --------------------------------------------------------------------------
    #
    def _synchronizer_work(self):
        '''
        **Purpose**: Thread in the master process to keep the workflow data
                     structure in appmanager up to date. We receive only tasks
                     objects from the task manager.

        Details:     Important to note that acknowledgements of the type
                     `channel.basic_ack()` is an acknowledgement to the server
                     that the msg was received.  This is not to be confused with
                     the Ack sent to the task_manager through the sync-ack
                     queue.
        '''

        self._prof.prof('sync_thread_start', uid=self._uid)
        self._logger.info('synchronizer thread started')

        while not self._terminate_sync.is_set():

            # wrapper to call `_update_task()`
            action  = self._get_message_to_sync('tmgr-to-sync')
            action += self._get_message_to_sync('cb-to-sync'  )

            if not action:
                time.sleep(0.01)

            # Raise an exception while running tests
            ru.raise_on(tag='sync_fail')

        self._prof.prof('sync_thread_stop', uid=self._uid)


# ------------------------------------------------------------------------------
# pylint: enable=protected-access





__copyright__ = 'Copyright 2017-2018, http://radical.rutgers.edu'
__author__    = 'Vivek Balasubramanian <vivek.balasubramanian@rutgers.edu>'
__license__   = 'MIT'

import queue

import threading       as mt
import multiprocessing as mp

from ...exceptions       import EnTKError
from ...                 import states, Task
from ..base.task_manager import Base_TaskManager


# pylint: disable=unused-argument
# ------------------------------------------------------------------------------
#
class TaskManager(Base_TaskManager):
    '''
    A Task Manager takes the responsibility of dispatching tasks it receives
    from a 'pending' queue for execution on to the available resources using a
    runtime system. Once the tasks have completed execution, they are pushed
    on to the completed queue for other components of EnTK to process.

    :arguments:
        :rmgr:              (ResourceManager) Object to be used to access the
                            Pilot where the tasks can be submitted

    Currently, EnTK is configured to work with one pending queue and one
    completed queue. In the future, the number of queues can be varied for
    different throughput requirements at the cost of additional Memory and CPU
    consumption.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, sid, rmgr, zmq_info):

        super().__init__(sid, rmgr, rts='mock', zmq_info=zmq_info)

        self._rts_runner = None
        self._zmq_info   = zmq_info

        self._log.info('Created task manager object: %s', self._uid)
        self._prof.prof('tmgr_create', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def _tmgr(self, uid, rmgr, zmq_info):
        '''
        **Purpose**: Method to be run by the tmgr process. This method receives
                     a Task from the 'pending' and submits it to the RTS.
                     Currently, it also converts Tasks into CUDs and CUs into
                     (partially described) Tasks.  This conversion is necessary
                     since the current RTS is RADICAL Pilot.  Once Tasks are
                     recovered from a CU, they are then pushed to the
                     completed queue. At all state transititons, they are synced
                     (blocking) with the AppManager in the master process.

        **Details**: The AppManager can re-invoke the tmgr process with this
                    function if the execution of the workflow is still
                    incomplete. There is also population of a dictionary,
                    `placeholders`, which stores the path of each of the tasks
                    on the remote machine.
        '''

        try:

            self._setup_zmq(zmq_info)

            self._prof.prof('tmgr process started', uid=self._uid)
            self._log.info('Task Manager process started')

            # Queue for communication between threads of this process
            task_queue = queue.Queue()

            # Start second thread to receive tasks and push to RTS
            self._rts_runner = mt.Thread(target=self._process_tasks,
                                         args=(task_queue, rmgr))
            self._rts_runner.daemon = True
            self._rts_runner.start()

            self._prof.prof('tmgr infrastructure setup done', uid=uid)

            # While we are supposed to run and the thread that does the work is
            # alive go.
            while not self._tmgr_terminate.is_set():

                try:

                    msgs = self._zmq_queue['get'].get_nowait(
                            qname='pending', timeout=100)

                    if msgs:

                        for msg in msgs:

                            if msg['type'] == 'workload':
                                task_queue.put(msg['body'])

                            elif msg['type'] == 'rts':
                                pass

                            else:
                                self._log.error('TMGR receiver wrong message type')

                except Exception as e:
                    self._log.exception('Error in task execution')
                    raise EnTKError(e) from e

            self._log.debug('Exited TMGR main loop')

        except KeyboardInterrupt:

            self._log.exception('Execution interrupted (probably by Ctrl+C), '
                                'cancel tmgr process gracefully...')
            raise


        except Exception as e:

            self._log.exception('task %s failed')
            raise EnTKError(e) from e

        finally:

            self._prof.prof('tmgr_term', uid=uid)

            if self._rts_runner:
                self._rts_runner.join()

            self._log.debug('TMGR RTS Runner joined')
            self._prof.close()
            self._log.debug('TMGR profile closed')


    # --------------------------------------------------------------------------
    #
    def _process_tasks(self, task_queue, rmgr):
        '''
        **Purpose**: The new thread that gets spawned by the main tmgr process
                     invokes this function. This function receives tasks from
                     'task_queue' and submits them to the RADICAL Pilot RTS.
        '''

        try:

            while not self._tmgr_terminate.is_set():

                body = None

                try:
                    body = task_queue.get_nowait()

                except queue.Empty:
                    # Ignore, we don't always have new tasks to run
                    pass

                if not body:
                    continue

                task_queue.task_done()

                bulk_tasks = list()

                for msg in body:

                    task = Task(from_dict=msg)
                    bulk_tasks.append(task)

                    self._advance(task, 'Task', states.SUBMITTING,
                                      'tmgr-to-sync')


                # this mock RTS immmedialtely completes all tasks
                for task in bulk_tasks:

                    task.exit_code = 0
                    self._advance(task, 'Task', states.COMPLETED, 'cb-to-sync')

                    self._log.info('Pushed task %s with state %s to completed',
                                   task.uid, task.state)

                    tdict = task.as_dict()
                    self._zmq_queue['put'].put(qname='completed', msgs=[tdict])

        except KeyboardInterrupt:
            self._log.exception('Execution interrupted (probably by Ctrl+C), '
                                'cancel task processor gracefully...')
            raise

        except Exception as e:
            self._log.exception('%s failed', self._uid)
            raise EnTKError(e) from e


    # --------------------------------------------------------------------------
    #
    def start_manager(self):
        '''
        **Purpose**: Method to start the tmgr process. The tmgr function
                     is not to be accessed directly. The function is started
                     in a separate thread using this method.
        '''
        # pylint: disable=attribute-defined-outside-init, access-member-before-definition
        if self._tmgr_process:
            self._log.warn('tmgr process already running!')
            return

        try:
            self._prof.prof('creating tmgr process', uid=self._uid)
            self._tmgr_terminate = mp.Event()

            self._tmgr_process = mp.Process(target=self._tmgr,
                                            name='task-manager',
                                            args=(self._uid,
                                                  self._rmgr,
                                                  self._zmq_info)
                                            )

            self._log.info('Starting task manager process')
            self._prof.prof('starting tmgr process', uid=self._uid)

            self._tmgr_process.start()
            self._log.debug('tmgr pid %s', self._tmgr_process.pid)

            return True

        except Exception as e:

            self._log.exception('Task manager not started')
            self.terminate_manager()
            raise EnTKError(e) from e


# ------------------------------------------------------------------------------
# pylint: enable=attribute-defined-outside-init, access-member-before-definition



#!/usr/bin/env python3

import radical.entk as re

#
# This example demonstrates the execution of a simple ensemble of simulations.
# Each ensemble member is in itself a pipeline of three different stages:
#
#     1) generate a random seed as input data
#     2) evolve a model based on that input data via a set of ensembles
#     3) derive a common metric across the model results
#
#   Similar patterns are frequently found in MD simulation workflows.  For the
#   purpose of this tutorial, the stages are:
#
#   - random seed  : create a random number)
#   - evolve model : N tasks computing n'th power of the input)
#   - common metric: sum over all 'model' outputs
#
# The final results are then staged back and printed on STDOUT.
#


# ------------------------------------------------------------------------------
#
def generate_pipeline(pname):
    '''
    Generate a single simulation pipeline, i.e., a new ensemble member.
    The pipeline structure consisting of three steps as described above.
    '''

    # all tasks in this pipeline share the same sandbox
    sandbox = pname

    # first stage: create 1 task to generate a random seed number
    t1 = re.Task()
    t1.executable = '/bin/sh'
    t1.arguments  = ['-c', 'od -An -N1 -i /dev/random']
    t1.stdout     = 'random.txt'
    t1.sandbox    = sandbox

    s1 = re.Stage()
    s1.add_tasks(t1)

    # second stage: create 10 tasks to compute the n'th power of that number
    s2 = re.Stage()
    n_simulations = 10
    for i in range(n_simulations):
        t2 = re.Task()
        t2.executable = '/bin/sh'
        t2.arguments  = ['-c', 'echo "$(cat random.txt) ^ %d" | bc' % i]
        t2.stdout     = 'power.%03d.txt' % i
        t2.sandbox    = sandbox
        s2.add_tasks(t2)

    # third stage: compute sum over all powers
    t3 = re.Task()
    t3.executable = '/bin/sh'
    t3.arguments  = ['-c', 'cat power.*.txt | paste -sd+ | bc']
    t3.stdout     = 'sum.txt'
    t3.sandbox    = sandbox

    s3 = re.Stage()
    s3.add_tasks(t3)

    # fourth stage: compute square root of previous sum
    t4 = re.Task()
    t4.executable = '/bin/sh'
    t4.arguments  = ['-c', 'echo "sqrt($(cat sum.txt))" | bc']
    t4.stdout     = 'sqrt.txt'
    t4.sandbox    = sandbox

    # download the result while renaming to get unique files per pipeline
    t4.download_output_data = ['sqrt.txt > %s.sqrt.txt' % pname]

    s4 = re.Stage()
    s4.add_tasks(t4)

    # assemble the three stages into a pipeline and return it
    p = re.Pipeline()
    p.add_stages(s1)
    p.add_stages(s2)
    p.add_stages(s3)
    p.add_stages(s4)

    return p


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # create application manager which executes our ensemble
    appman = re.AppManager()

    # assign resource request description to the application manager using
    # three mandatory keys: target resource, walltime, and number of cpus
    appman.resource_desc = {
        'resource': 'local.localhost',
      # 'resource': 'local.localhost_flux',
        'walltime': 10,
        'cpus'    : 1
    }

    # create an ensemble of n simulation pipelines
    n_pipelines = 10
    ensemble = set()
    for cnt in range(n_pipelines):
        ensemble.add(generate_pipeline(pname='pipe.%03d' % cnt))

    # assign the workflow to the application manager, then
    # run the ensemble and wait for completion
    appman.workflow = ensemble
    appman.run()

    # check results which were staged back
    for cnt in range(n_pipelines):
        pipe_fname  = 'pipe.%03d.sqrt.txt' % cnt
        pipe_result = float(open(pipe_fname, encoding='utf-8').read())
        print('%3d -- %25.2f' % (cnt, pipe_result))


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2017-2018, http://radical.rutgers.edu'
__author__    = 'Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>'
__license__   = 'MIT'

# pylint: disable=unused-argument

import radical.utils as ru

from ..base.resource_manager import Base_ResourceManager


# ------------------------------------------------------------------------------
#
class ResourceManager(Base_ResourceManager):
    '''
    A resource manager takes the responsibility of placing resource requests on
    different, possibly multiple, DCIs. This ResourceManager uses mocks an
    implementation by doing nothing, it is only usable for testing.

    :arguments:
        :resource_desc: dictionary with details of the resource request and
                        access credentials of the user
        :example: resource_desc = {
                                    |  'resource'      : 'xsede.stampede',
                                    |  'walltime'      : 120,
                                    |  'cpus'          : 64,
                                    |  'project'       : 'TG-abcxyz',
                                    |  'queue'         : 'abc',    # optional
                                    |  'access_schema' : 'ssh'     # optional}
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, resource_desc, sid, rts_config):

        super(ResourceManager, self).__init__(resource_desc=resource_desc,
                                              sid=sid,
                                              rts='mock',
                                              rts_config=rts_config)


    # --------------------------------------------------------------------------
    #
    def get_resource_allocation_state(self):
        '''
        **Purpose**: get the state of the resource allocation

        '''
        try:
            ru.raise_on(tag='resource_fail')
            return 'Running'
        except:
            return 'Final'


    # --------------------------------------------------------------------------
    #
    def get_completed_states(self):
        '''
        **Purpose**: test if a resource allocation was submitted
        '''

        return ['Final']


    # --------------------------------------------------------------------------
    #
    def _validate_resource_desc(self):
        '''
        **Purpose**: validate the provided resource description
        '''

        return True


    # --------------------------------------------------------------------------
    #
    def _populate(self):
        '''
        **Purpose**: evaluate attributes provided in the resource description
        '''

        return None


    # --------------------------------------------------------------------------
    #
    def submit_resource_request(self, *args):
        '''
        **Purpose**: Create a resourceas per provided resource description
        '''

        return None


    # --------------------------------------------------------------------------
    #
    def get_rts_info(self):

        return None


    # --------------------------------------------------------------------------
    #
    def _terminate_resource_request(self):
        '''
        **Purpose**: Cancel the resource
        '''

        return None


# ------------------------------------------------------------------------------

